# -*- coding: utf-8 -*-

import argparse
import logging
import sys

import yaml

from mau import __version__, Mau

__author__ = "Leonardo Giordani"
__copyright__ = "Leonardo Giordani"
__license__ = "mit"

_logger = logging.getLogger(__name__)


def parse_args(args):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--input-file",
        action="store",
        required=True,
        help="Input file",
    )

    parser.add_argument(
        "-o",
        "--output-file",
        action="store",
        required=False,
        help="Optional output file",
    )

    parser.add_argument(
        "-c",
        "--config-file",
        action="store",
        required=False,
        help="Optional config file",
    )

    parser.add_argument(
        "-f",
        "--format",
        action="store",
        required=False,
        default="html",
        choices=["html", "asciidoctor"],
        help="Output format",
    )

    parser.add_argument(
        "--verbose",
        dest="loglevel",
        help="set loglevel to INFO",
        action="store_const",
        const=logging.INFO,
    )

    parser.add_argument(
        "--debug",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        const=logging.DEBUG,
    )

    parser.add_argument(
        "--version", action="version", version=f"Mau version {__version__}"
    )

    return parser.parse_args()


def setup_logging(loglevel):
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(
        level=loglevel, stream=sys.stdout, format=logformat, datefmt="%Y-%m-%d %H:%M:%S"
    )


def main(args):
    args = parse_args(args)
    setup_logging(args.loglevel)

    with open(args.input_file, "r") as f:
        text = f.read()

    config = {}
    if args.config_file:
        with open(args.config_file, "r") as f:
            config = yaml.load(f, Loader=yaml.FullLoader)

    templates_directory = f'{config["theme"]}/templates' if "theme" in config else None

    mau = Mau(
        config, target_format=args.format, templates_directory=templates_directory
    )
    output = mau.process(text)

    if args.format == "html":
        output_extension = ".html"
    elif args.format == "asciidoctor":
        output_extension = ".adoc"

    output_file = args.output_file or args.input_file.replace(".mau", output_extension)

    with open(output_file, "w") as f:
        text = f.write(output)


def run():
    main(sys.argv[1:])


if __name__ == "__main__":
    run()



import textwrap

from mau.lexers.base_lexer import Text, Literal, EOL, EOF, WS
from mau.lexers.main_lexer import MainLexer


def dedent(text):
    return textwrap.dedent(text).strip()


def test_empty_text():
    lex = MainLexer()

    lex.process("")

    assert lex.tokens == [EOL, EOF]


def test_empty_lines():
    lex = MainLexer()

    lex.process("\n")

    assert lex.tokens == [EOL, EOL, EOF]


def test_lines_with_only_spaces():
    lex = MainLexer()

    lex.process("      \n      ")

    assert lex.tokens == [EOL, EOL, EOF]


def test_horizontal_rule():
    lex = MainLexer()

    lex.process("---")

    assert lex.tokens == [Literal("---"), EOL, EOF]


def test_attributes():
    lex = MainLexer()

    lex.process("[name]")

    assert lex.tokens == [
        Literal("["),
        Text("name"),
        Literal("]"),
        EOL,
        EOF,
    ]


def test_attributes_marker_in_text():
    lex = MainLexer()

    lex.process("Not [attributes]")

    assert lex.tokens == [
        Text("Not [attributes]"),
        EOL,
        EOF,
    ]


def test_variable_definition():
    lex = MainLexer()

    lex.process(":variable:value123")

    assert lex.tokens == [
        Literal(":"),
        Text("variable"),
        Literal(":"),
        Text("value123"),
        EOL,
        EOF,
    ]


def test_variable_negation():
    lex = MainLexer()

    lex.process(":!variable:")

    assert lex.tokens == [
        Literal(":"),
        Text("!variable"),
        Literal(":"),
        EOL,
        EOF,
    ]


def test_variable_marker_in_text():
    lex = MainLexer()

    lex.process("Not a :variable:")

    assert lex.tokens == [
        Text("Not a :variable:"),
        EOL,
        EOF,
    ]


def test_variable_definition_accepted_characters():
    lex = MainLexer()

    lex.process(":abcAB.C0123-_:value123")

    assert lex.tokens == [
        Literal(":"),
        Text("abcAB.C0123-_"),
        Literal(":"),
        Text("value123"),
        EOL,
        EOF,
    ]


def test_multiple_lines():
    lex = MainLexer()

    lex.process(
        dedent(
            """
            This is text
            split into multiple lines

            with an empty line
            """
        )
    )

    assert lex.tokens == [
        Text("This is text"),
        EOL,
        Text("split into multiple lines"),
        EOL,
        EOL,
        Text("with an empty line"),
        EOL,
        EOF,
    ]


def test_multiple_lines_poistions():
    lex = MainLexer()

    lex.process(
        dedent(
            """
            This is text
            split into multiple lines

            with an empty line
            """
        )
    )

    assert [i.position for i in lex.tokens] == [
        (0, 0),
        (0, 12),
        (1, 0),
        (1, 25),
        (2, 0),
        (3, 0),
        (3, 18),
        (4, 0),
    ]


def test_title():
    lex = MainLexer()

    lex.process(
        dedent(
            """
            .A title
            Some text
            """
        )
    )

    assert lex.tokens == [
        Literal("."),
        Text("A title"),
        EOL,
        Text("Some text"),
        EOL,
        EOF,
    ]


def test_title_with_space():
    lex = MainLexer()

    lex.process(
        dedent(
            """
            . A title
            Some text
            """
        )
    )

    assert lex.tokens == [
        Literal("."),
        WS(" "),
        Text("A title"),
        EOL,
        Text("Some text"),
        EOL,
        EOF,
    ]


def test_title_multiple_spaces_after_mark():
    lex = MainLexer()

    lex.process(
        dedent(
            """
            .     A title with spaces
            Some text
            """
        )
    )

    assert lex.tokens == [
        Literal("."),
        WS("     "),
        Text("A title with spaces"),
        EOL,
        Text("Some text"),
        EOL,
        EOF,
    ]


def test_id():
    lex = MainLexer()

    lex.process(
        dedent(
            """
            #someid
            Some text
            """
        )
    )

    assert lex.tokens == [
        Literal("#"),
        Text("someid"),
        EOL,
        Text("Some text"),
        EOL,
        EOF,
    ]


def test_command():
    lex = MainLexer()

    lex.process("::command:arg0,arg1")

    assert lex.tokens == [
        Literal("::"),
        Text("command"),
        Literal(":"),
        Text("arg0,arg1"),
        EOL,
        EOF,
    ]


def test_command_without_arguments():
    lex = MainLexer()

    lex.process("::command:")

    assert lex.tokens == [
        Literal("::"),
        Text("command"),
        Literal(":"),
        EOL,
        EOF,
    ]


def test_comment():
    lex = MainLexer()

    lex.process("// Some comment")

    assert lex.tokens == [
        Text("// Some comment"),
        EOL,
        EOF,
    ]


def test_multiline_comment():
    lex = MainLexer()

    lex.process(
        dedent(
            """
            ////
            Some comment

               another line
            ////
            """
        )
    )

    assert lex.tokens == [
        Literal("////"),
        EOL,
        Text("Some comment"),
        EOL,
        EOL,
        Text("   another line"),
        EOL,
        Literal("////"),
        EOL,
        EOF,
    ]


def test_include_content():
    lex = MainLexer()

    lex.process("<< type:/path/to/it.jpg")

    assert lex.tokens == [
        Literal("<<"),
        WS(" "),
        Text("type:/path/to/it.jpg"),
        EOL,
        EOF,
    ]


def test_include_content_multiple_spaces_after_mark():
    lex = MainLexer()

    lex.process("<<      type:/path/to/it.jpg")

    assert lex.tokens == [
        Literal("<<"),
        WS("      "),
        Text("type:/path/to/it.jpg"),
        EOL,
        EOF,
    ]


def test_include_content_positions():
    lex = MainLexer()

    lex.process("<< type:/path/to/it.jpg")

    assert [i.position for i in lex.tokens] == [
        (0, 0),
        (0, 2),
        (0, 3),
        (0, 23),
        (1, 0),
    ]


def test_include_content_with_arguments():
    lex = MainLexer()

    lex.process("<< type:/path/to/it.jpg(value1,argument2=value2)")

    assert lex.tokens == [
        Literal("<<"),
        WS(" "),
        Text("type:/path/to/it.jpg"),
        Literal("("),
        Text("value1,argument2=value2"),
        Literal(")"),
        EOL,
        EOF,
    ]


def test_unordered_list():
    lex = MainLexer()

    lex.process("* Item")

    assert lex.tokens == [
        Literal("*"),
        WS(" "),
        Text("Item"),
        EOL,
        EOF,
    ]


def test_unordered_list_leading_space():
    lex = MainLexer()

    lex.process("    * Item")

    assert lex.tokens == [
        WS("    "),
        Literal("*"),
        WS(" "),
        Text("Item"),
        EOL,
        EOF,
    ]


def test_unordered_list_trailing_space():
    lex = MainLexer()

    lex.process("*       Item")

    assert lex.tokens == [
        Literal("*"),
        WS("       "),
        Text("Item"),
        EOL,
        EOF,
    ]


def test_unordered_list_multiple_stars():
    lex = MainLexer()

    lex.process("*** Item")

    assert lex.tokens == [
        Literal("***"),
        WS(" "),
        Text("Item"),
        EOL,
        EOF,
    ]


def test_ordered_list():
    lex = MainLexer()

    lex.process("# Item")

    assert lex.tokens == [
        Literal("#"),
        WS(" "),
        Text("Item"),
        EOL,
        EOF,
    ]


def test_ordered_list_multiple_stars():
    lex = MainLexer()

    lex.process("### Item")

    assert lex.tokens == [
        Literal("###"),
        WS(" "),
        Text("Item"),
        EOL,
        EOF,
    ]


def test_header():
    lex = MainLexer()

    lex.process("= Header")

    assert lex.tokens == [
        Literal("="),
        WS(" "),
        Text("Header"),
        EOL,
        EOF,
    ]


def test_header_multiple_spaces_after_mark():
    lex = MainLexer()

    lex.process("=    Header")

    assert lex.tokens == [
        Literal("="),
        WS("    "),
        Text("Header"),
        EOL,
        EOF,
    ]


def test_empty_header():
    lex = MainLexer()

    lex.process("=")

    assert lex.tokens == [
        Text("="),
        EOL,
        EOF,
    ]


def test_multiple_header_markers():
    lex = MainLexer()

    lex.process("=== Header")

    assert lex.tokens == [
        Literal("==="),
        WS(" "),
        Text("Header"),
        EOL,
        EOF,
    ]


def test_header_marker_in_header_text():
    lex = MainLexer()

    lex.process("= a=b")

    assert lex.tokens == [
        Literal("="),
        WS(" "),
        Text("a=b"),
        EOL,
        EOF,
    ]


def test_header_markers_in_text():
    lex = MainLexer()

    lex.process("Definitely not a === header")

    assert lex.tokens == [
        Text("Definitely not a === header"),
        EOL,
        EOF,
    ]


def test_unlisted_header():
    lex = MainLexer()

    lex.process("==! Header")

    assert lex.tokens == [
        Literal("==!"),
        WS(" "),
        Text("Header"),
        EOL,
        EOF,
    ]



import copy
import logging

import jinja2

_logger = logging.getLogger(__name__)


class TemplateNotFound(ValueError):
    pass


class Visitor:
    _template_extension = ".txt"
    _environment_options = {}

    def __init__(
        self,
        default_templates=None,
        templates_directory=None,
        config=None,
        toc=None,
        footnotes=None,
    ):
        self.default_templates = default_templates or {}
        self.templates_directory = templates_directory
        self.config = copy.deepcopy(config) if config else {}
        self.toc = toc or []
        self.footnotes = footnotes or []

        self.default_templates_env = jinja2.Environment(
            loader=jinja2.DictLoader(self.default_templates),
            **self._environment_options,
        )

        if self.templates_directory:
            self.main_templates_env = jinja2.Environment(
                loader=jinja2.FileSystemLoader(searchpath=self.templates_directory),
                **self._environment_options,
            )
        else:
            self.main_templates_env = self.default_templates_env

    def _template(self, node_type):  # pragma: no cover
        template_name = f"{node_type}{self._template_extension}"

        try:
            return self.main_templates_env.get_template(template_name)
        except jinja2.exceptions.TemplateNotFound:
            _logger.debug(
                f"Template {template_name} not provided, trying the default one"
            )
            try:
                return self.default_templates_env.get_template(template_name)
            except jinja2.exceptions.TemplateNotFound as e:
                raise TemplateNotFound(e)

    def _render(self, node_type, **kwargs):
        template = self._template(node_type)

        return template.render(config=self.config, **kwargs)

    def visit(self, node, *args, **kwargs):
        if node is None:
            return ""

        keys = getattr(self, f'_visit_{node["type"]}')(node, *args, **kwargs)

        node_types = []
        node_types.extend(keys.pop("node_types", []))
        node_types.append(node["type"])

        for i in node_types:
            try:
                return self._render(i, **keys)
            except TemplateNotFound:
                continue

        raise ValueError(f"Cannot find any suitable template among {node_types}")

    def visitlist(self, nodes):
        return [self.visit(i) for i in nodes]



from mau.lexers.base_lexer import Token, TokenTypes
from mau.lexers.arguments_lexer import ArgumentsLexer
from mau.parsers.base_parser import BaseParser, ParseError


def merge_args(args, kwargs, names):
    named_args = {}

    for i, name in enumerate(names):
        try:
            named_args[name] = args[i]
        except IndexError:
            pass

    named_args.update(kwargs)

    return [], named_args


class ArgumentsParser(BaseParser):
    def __init__(self, raw=False):
        super().__init__()
        self.lexer = ArgumentsLexer()
        self.raw = raw
        self._named_arguments = False
        self.kwargs = {}
        self.args = []

    def _parse_named_argument(self):
        name = self.get_token(TokenTypes.TEXT).value
        self.get_token(TokenTypes.LITERAL, "=")

        if self.peek_token_is(TokenTypes.LITERAL, '"'):
            self.get_token(TokenTypes.LITERAL, '"')
            value = self.collect_join([Token(TokenTypes.LITERAL, '"')])
            self.get_token(TokenTypes.LITERAL, '"')
        else:
            value = self.get_token(TokenTypes.TEXT).value

        return name, value

    def _parse_unnamed_argument(self):
        if self.peek_token_is(TokenTypes.LITERAL, '"'):
            self.get_token(TokenTypes.LITERAL, '"')
            value = self.collect_join([Token(TokenTypes.LITERAL, '"')])
            self.get_token(TokenTypes.LITERAL, '"')
        else:
            value = self.get_token(TokenTypes.TEXT).value

        return value

    def _parse_single_argument(self):
        if self.raw:
            value = self.collect_join([Token(TokenTypes.EOL), Token(TokenTypes.EOF)])
            self.args.append(value)
            return

        with self:
            name, value = self._parse_named_argument()
            self.kwargs[name] = value
            self._named_arguments = True
            return

        if self._named_arguments:
            raise ParseError("Unnamed arguments after named arguments are forbidden")

        with self:
            value = self._parse_unnamed_argument()
            self.args.append(value)
            return

    def _parse_arguments(self):
        with self:
            while True:
                self._parse_single_argument()

                with self:
                    self.get_token(TokenTypes.LITERAL, ",")
                    with self:
                        self.get_token(TokenTypes.WHITESPACE)
                    continue

                break

    def _parse(self):
        self._parse_arguments()



import re

from mau.lexers.base_lexer import BaseLexer, TokenTypes
from mau.lexers.arguments_lexer import ArgumentsLexer


class MainLexer(BaseLexer):
    def _process_empty_line(self):
        regexp = re.compile(r"^\ +$")

        match = regexp.match(self._current_line)

        if not match:
            return None

        self._skip(len(match.group()))

    def _process_horizontal_rule(self):
        if not self._current_line == "---":
            return None

        return self._create_token_and_skip(TokenTypes.LITERAL, self._current_line)

    def _process_comment(self):
        regexp = re.compile("^(//.*)")

        match = regexp.match(self._current_line)

        if not match:
            return None

        tokens = [
            self._create_token_and_skip(TokenTypes.TEXT, match.group(1)),
            self._create_token_and_skip(TokenTypes.EOL),
        ]

        return tokens

    def _process_multiline_comment(self):
        if not self._current_line.startswith("////"):
            return None

        tokens = [
            self._create_token_and_skip(TokenTypes.LITERAL, self._current_line),
            self._create_token_and_skip(TokenTypes.EOL),
        ]
        return tokens

    def _process_header(self):
        regexp = re.compile("^(=+!?)( +)(.*)")

        match = regexp.match(self._current_line)

        if not match:
            return None

        if not match.group(2):
            return None

        if not match.group(3):
            return None

        tokens = [
            self._create_token_and_skip(TokenTypes.LITERAL, match.group(1)),
            self._create_token_and_skip(TokenTypes.WHITESPACE, match.group(2)),
            self._create_token_and_skip(TokenTypes.TEXT, match.group(3)),
            self._create_token_and_skip(TokenTypes.EOL),
        ]

        return tokens

    def _process_list(self):
        regexp = re.compile(r"^( *)([\*#]+)( +)(.*)")

        match = regexp.match(self._current_line)

        if not match:
            return None

        if not match.group(4):
            return None

        tokens = []

        if match.group(1):
            tokens.append(
                self._create_token_and_skip(TokenTypes.WHITESPACE, match.group(1))
            )

        tokens.extend(
            [
                self._create_token_and_skip(TokenTypes.LITERAL, match.group(2)),
                self._create_token_and_skip(TokenTypes.WHITESPACE, match.group(3)),
                self._create_token_and_skip(TokenTypes.TEXT, match.group(4)),
            ]
        )

        return tokens

    def _process_id(self):
        if self._current_char != "#":
            return None

        return self._create_token_and_skip(TokenTypes.LITERAL, self._current_char)

    def _process_title(self):
        regexp = re.compile(r"^\.( *)(.*)")

        match = regexp.match(self._current_line)

        if not match:
            return None

        if not match.group(2):
            return None

        tokens = [
            self._create_token_and_skip(TokenTypes.LITERAL, "."),
        ]

        if match.group(1):
            tokens.append(
                self._create_token_and_skip(TokenTypes.WHITESPACE, match.group(1))
            )

        tokens.extend(
            [
                self._create_token_and_skip(TokenTypes.TEXT, match.group(2)),
                self._create_token_and_skip(TokenTypes.EOL),
            ]
        )

        return tokens

    def _process_attributes(self):
        if self._current_char != "[":
            return None

        if not self._current_line.endswith("]"):
            return None

        tokens = [self._create_token_and_skip(TokenTypes.LITERAL, "[")]

        line = self._current_line[1:-1]

        tokens.append(self._create_token_and_skip(TokenTypes.TEXT, line))
        tokens.append(self._create_token_and_skip(TokenTypes.LITERAL, "]"))
        tokens.append(self._create_token_and_skip(TokenTypes.EOL))

        return tokens

    def _process_include_content(self):
        regexp = re.compile(r"^<<( +)([^(]+)((.*))?")

        match = regexp.match(self._current_line)

        if not match:
            return None

        if not match.group(2):
            return None

        tokens = [
            self._create_token_and_skip(TokenTypes.LITERAL, "<<"),
            self._create_token_and_skip(TokenTypes.WHITESPACE, match.group(1)),
            self._create_token_and_skip(TokenTypes.TEXT, match.group(2)),
        ]

        if match.group(3):
            tokens.extend(
                [
                    self._create_token_and_skip(TokenTypes.LITERAL, "("),
                    self._create_token_and_skip(TokenTypes.TEXT, match.group(3)[1:-1]),
                    self._create_token_and_skip(TokenTypes.LITERAL, ")"),
                ]
            )

        tokens.append(self._create_token_and_skip(TokenTypes.EOL))
        return tokens

    def _process_command(self):
        if not self._current_line.startswith("::"):
            return None

        regexp = re.compile(r"::([a-z0-9_]+):(.*)?")

        match = regexp.match(self._current_line)

        if not match:  # pragma: no cover
            return None

        tokens = [
            self._create_token_and_skip(TokenTypes.LITERAL, "::"),
            self._create_token_and_skip(TokenTypes.TEXT, match.group(1)),
            self._create_token_and_skip(TokenTypes.LITERAL, ":"),
        ]

        if match.group(2):
            tokens.append(self._create_token_and_skip(TokenTypes.TEXT, match.group(2)))

        tokens.append(self._create_token_and_skip(TokenTypes.EOL))

        return tokens

    def _process_variable_definition(self):
        if self._current_char != ":":
            return None

        regexp = re.compile(r":([!a-zA-Z0-9\-_\.]+):(.*)?")

        match = regexp.match(self._current_line)

        if not match:  # pragma: no cover
            return None

        tokens = [
            self._create_token_and_skip(TokenTypes.LITERAL, ":"),
            self._create_token_and_skip(TokenTypes.TEXT, match.group(1)),
            self._create_token_and_skip(TokenTypes.LITERAL, ":"),
        ]

        if match.group(2):
            tokens.append(
                self._create_token_and_skip(TokenTypes.TEXT, (match.group(2)))
            )

        tokens.append(self._create_token_and_skip(TokenTypes.EOL))

        return tokens

    def _process_text(self):
        tokens = [
            self._create_token_and_skip(TokenTypes.TEXT, self._tail),
            self._create_token_and_skip(TokenTypes.EOL),
        ]
        return tokens

    def _process_functions(self):
        return [
            self._process_eof,
            self._process_empty_line,
            self._process_eol,
            self._process_horizontal_rule,
            self._process_multiline_comment,
            self._process_comment,
            self._process_command,
            self._process_header,
            self._process_list,
            self._process_title,
            self._process_id,
            self._process_attributes,
            self._process_include_content,
            self._process_variable_definition,
            self._process_text,
        ]



from mau.visitors.html_visitor import HTMLVisitor

from mau.parsers.main_parser import MainParser
from tests.helpers import remove_indentation, init_parser_factory, visitlist_factory

init_parser = init_parser_factory(MainParser)
visitlist = visitlist_factory(HTMLVisitor)


def test_toc_entry():
    v = HTMLVisitor()

    result = v._visit_toc_entry(
        {
            "type": "toc_entry",
            "level": 1,
            "text": "Header 1",
            "anchor": "Header 1-XXXXXX",
            "children": [],
        }
    )

    assert result == {
        "text": "Header 1",
        "anchor": "Header 1-XXXXXX",
        "children": "",
    }


def test_command_toc():

    toc = [
        {
            "type": "toc_entry",
            "level": 1,
            "text": "Header 1",
            "anchor": "Header 1-XXXXXX",
            "children": [
                {
                    "type": "toc_entry",
                    "level": 2,
                    "text": "Header 1.1",
                    "anchor": "Header 1.1-XXXXXX",
                    "children": [],
                },
            ],
        },
        {
            "type": "toc_entry",
            "level": 1,
            "text": "Header 2",
            "anchor": "Header 2-XXXXXX",
            "children": [],
        },
    ]

    parser = init_parser("::toc:")
    parser.parse()

    result = visitlist([node.asdict() for node in parser.nodes], toc=toc)

    assert result == [
        remove_indentation(
            """
            <div>
              <ul>
                <li>
                  <a href="#Header 1-XXXXXX">Header 1</a>
                  <ul>
                    <li>
                      <a href="#Header 1.1-XXXXXX">Header 1.1</a>
                    </li>
                  </ul>
                </li>
                <li>
                  <a href="#Header 2-XXXXXX">Header 2</a>
                </li>
              </ul>
            </div>
            """
        )
    ]



import pytest

from mau.parsers.preprocess_variables_parser import (
    PreprocessVariablesParser,
    PreprocessError,
)

from tests.helpers import listasdict, init_parser_factory, dedent

init_parser = init_parser_factory(PreprocessVariablesParser)


def test_plain_text():
    p = init_parser("This is text", {"attr": "5"})
    p.parse()

    assert listasdict(p.nodes) == [
        {
            "type": "text",
            "value": "This is text\n",
        }
    ]


def test_replace_variable():
    p = init_parser("This is number {attr}", {"attr": "5"})
    p.parse()

    assert listasdict(p.nodes) == [
        {
            "type": "text",
            "value": "This is number 5\n",
        }
    ]


def test_replace_variable_with_namespace():
    p = init_parser("This is number {config.attr}", {"config": {"attr": "5"}})
    p.parse()

    assert listasdict(p.nodes) == [
        {
            "type": "text",
            "value": "This is number 5\n",
        }
    ]


def test_replace_boolean():
    p = init_parser("This flag is {flag}", {"flag": True})
    p.parse()

    assert listasdict(p.nodes) == [
        {
            "type": "text",
            "value": "This flag is True\n",
        }
    ]


def test_escape_curly_braces():
    p = init_parser(r"This is \{attr\}", {"attr": "5"})
    p.parse()

    assert listasdict(p.nodes) == [
        {
            "type": "text",
            "value": "This is {attr}\n",
        }
    ]


def test_curly_braces_in_verbatim():
    p = init_parser("This is `{attr}`", {"attr": "5"})
    p.parse()

    assert listasdict(p.nodes) == [
        {
            "type": "text",
            "value": "This is `{attr}`\n",
        }
    ]


def test_open_verbatim():
    p = init_parser("This is `{attr}", {"attr": "5"})
    p.parse()

    assert listasdict(p.nodes) == [
        {
            "type": "text",
            "value": "This is `5\n",
        }
    ]


def test_open_curly():
    p = init_parser("This is {attr", {"attr": "5"})
    p.parse()

    assert listasdict(p.nodes) == [
        {
            "type": "text",
            "value": "This is {attr\n",
        }
    ]


def test_escape_curly_braces_in_verbatim():
    p = init_parser(r"This is `\{attr\}`", {"attr": "5"})
    p.parse()

    assert listasdict(p.nodes) == [
        {
            "type": "text",
            "value": "This is `\\{attr\\}`\n",
        }
    ]


def test_escape_other_chars():
    p = init_parser(r"This \_is\_ \text", {"attr": "5"})
    p.parse()

    assert listasdict(p.nodes) == [
        {
            "type": "text",
            "value": "This \\_is\\_ \\text\n",
        }
    ]


def test_curly_braces_in_escaped_verbatim():
    p = init_parser(r"This is \`{attr}\`", {"attr": "5"})
    p.parse()

    assert listasdict(p.nodes) == [
        {
            "type": "text",
            "value": "This is \\`5\\`\n",
        }
    ]


def test_variable_not_existing():
    p = init_parser("This is number {attr}", {})

    with pytest.raises(PreprocessError):
        p.parse()


def test_variables_can_contain_markers():
    p = init_parser(
        dedent(
            """
            A very {bold} text
            Some code: {dictdef}
            """
        ),
        {"bold": "*bold*", "dictdef": "`adict = {'a':5}`"},
    )
    p.parse()

    assert listasdict(p.nodes) == [
        {
            "type": "text",
            "value": "A very *bold* text\nSome code: `adict = {'a':5}`\n",
        }
    ]


def test_escape_backtick():
    p = init_parser(r"This is `\``")
    p.parse()

    assert listasdict(p.nodes) == [
        {
            "type": "text",
            "value": "This is `\``\n",
        }
    ]



__config_version__ = 1

GLOBALS = {
    "serializer": "{{major}}.{{minor}}.{{patch}}",
}

FILES = [{"path": "setup.cfg", "serializer": "version = {{major}}.{{minor}}.{{patch}}"}]

VERSION = ["major", "minor", "patch"]

VCS = {
    "name": "git",
    "commit_message": (
        "Version updated from {{ current_version }}" " to {{ new_version }}"
    ),
    "target_branch": "main",
}



from mau.lexers.base_lexer import Token


def test_token_accepts_type_and_value():
    t = Token("sometype", "somevalue")

    assert t.type == "sometype"
    assert t.value == "somevalue"


def test_token_transforms_value_in_string():
    t = Token("sometype", 3)

    assert t.type == "sometype"
    assert t.value == "3"


def test_token_keeps_value_none():
    t = Token("sometype", None)

    assert t.type == "sometype"
    assert t.value is None


def test_token_transforms_zero():
    t = Token("sometype", 0)

    assert t.type == "sometype"
    assert t.value == "0"


def test_token_value_defaults_to_none():
    t = Token("sometype")

    assert t.type == "sometype"
    assert t.value is None


def test_token_string_representation():
    t = Token("sometype", "somevalue")

    assert str(t) == "Token(sometype, 'somevalue')"


def test_token_representation():
    t = Token("sometype", "somevalue")

    assert repr(t) == "Token(sometype, 'somevalue')"


def test_token_equality():
    assert Token("sometype", "somevalue") == Token("sometype", "somevalue")


def test_token_length():
    t = Token("sometype", "somevalue")

    assert len(t) == len("somevalue")
    assert bool(t) is True


def test_empty_token_has_length_zero():
    t = Token("sometype")

    assert len(t) == 0
    assert bool(t) is True


def test_token_accepts_text_position():
    line = 456
    column = 123
    t = Token("sometype", "somevalue", position=(line, column))

    assert t.type == "sometype"
    assert t.value == "somevalue"
    assert t.position == (line, column)


def test_token_equality_ignores_position():
    assert Token("sometype", "somevalue", position=(12, 34)) == Token(
        "sometype", "somevalue"
    )
    assert Token("sometype", "somevalue") == Token(
        "sometype", "somevalue", position=(12, 34)
    )


def test_token_equality_accepts_none():
    assert Token("sometype", "somevalue") is not None


def test_token_string_representation_with_position():
    t = Token("sometype", "somevalue", position=(12, 34))

    assert str(t) == "Token(sometype, 'somevalue', line=12, col=34)"


def test_token_equality_with_any():
    assert Token("sometype", "somevalue") == Token("sometype")



from pygments import highlight
from pygments.lexers import get_lexer_by_name
from pygments.formatters import get_formatter_by_name

from mau.visitors.visitor import Visitor

DEFAULT_TEMPLATES = {
    "admonition.html": (
        '<div class="admonition {{ class }}">'
        '<i class="{{ icon }}"></i>'
        '<div class="content">'
        '<div class="title">{{ label }}</div>'
        "<div>{{ content }}</div>"
        "</div></div>"
    ),
    "block.html": (
        '<div{% if type %} class="{{ type }}"{% endif %}>'
        '{% if title %}<div class="title">{{ title }}</div>{% endif %}'
        '<div class="content">{{ content }}</div>'
        "</div>"
    ),
    "callout.html": '<span class="callout">{{ name }}</span>',
    "class.html": '<span class="{{ classes }}">{{ content }}</span>',
    "command.html": "{{ content }}",
    "container.html": "{{ content }}",
    "document.html": "<html><head></head><body>{{ content }}</body></html>",
    "footnote_def.html": (
        '<div id="{{ defanchor }}">'
        '<a href="#{{ refanchor }}">{{ number }}</a> {{ text }}</div>'
    ),
    "footnote_ref.html": (
        "<sup>"
        '[<a id="{{ refanchor }}" href="#{{ defanchor }}">{{ number }}</a>]'
        "</sup>"
    ),
    "footnotes.html": '<div id="_footnotes">{{ entries }}</div>',
    "header.html": (
        '<h{{ level }} id="{{ anchor }}">'
        '{% if id %}<a id="{{ id }}"></a>{% endif %}'
        "{{ value }}</h{{ level }}>"
    ),
    "horizontal_rule.html": "<hr>",
    "image.html": (
        '<div class="imageblock">'
        '<div class="content">'
        '<img src="{{ uri }}"{% if alt_text %} alt="{{ alt_text }}"{% endif %}>'
        '{% if title %}<div class="title">{{ title }}</div>{% endif %}'
        "</div></div>"
    ),
    "inline_image.html": (
        '<span class="image">'
        '<img src="{{ uri }}"{%if alt_text %} alt="{{ alt_text }}"{% endif %}>'
        "</span>"
    ),
    "link.html": '<a href="{{ target }}">{{ text }}</a>',
    "list.html": (
        "<{% if ordered %}ol{% else %}ul{% endif %}>"
        "{{ items }}"
        "</{% if ordered %}ol{% else %}ul{% endif %}>"
    ),
    "list_item.html": "<li>{{ content }}</li>",
    "macro.html": "",
    "paragraph.html": "<p>{{ content }}</p>",
    "quote.html": (
        "<blockquote>" "{{ content }}" "<cite>{{ attribution }}</cite>" "</blockquote>"
    ),
    "raw.html": "{{ content }}",
    "sentence.html": "{{ content }}",
    "source.html": (
        '<div class="code">'
        '{% if title %}<div class="title">{{ title }}</div>{% endif %}'
        '<div class="content">{{ code }}</div>'
        '{% if callouts %}<div class="callouts">'
        "<table><tbody>"
        "{% for callout in callouts %}<tr>"
        "<td>{{ callout[0] }}</td>"
        "<td>{{ callout[1] }}</td>"
        "</tr>{% endfor %}"
        "</tbody></table>"
        "</div>{% endif %}"
        "</div>"
    ),
    "star.html": "<strong>{{ content }}</strong>",
    "text.html": "{{ value }}",
    "toc.html": "<div>{% if entries%}<ul>{{ entries }}</ul>{% endif %}</div>",
    "toc_entry.html": (
        "<li>"
        '<a href="#{{ anchor }}">{{ text }}</a>'
        "{% if children %}<ul>{{ children }}</ul>{% endif %}"
        "</li>"
    ),
    "underscore.html": "<em>{{ content }}</em>",
    "verbatim.html": "<code>{{ content }}</code>",
}


class HTMLVisitor(Visitor):
    _template_extension = ".html"

    def __init__(
        self,
        default_templates=None,
        templates_directory=None,
        config=None,
        toc=None,
        footnotes=None,
    ):
        super().__init__(
            default_templates=DEFAULT_TEMPLATES,
            templates_directory=templates_directory,
            config=config,
            toc=toc,
            footnotes=footnotes,
        )

    def _visit_text(self, node):
        return {"value": node["value"]}

    def _visit_sentence(self, node):
        return {"content": "".join([self.visit(t) for t in node["content"]])}

    def _visit_paragraph(self, node):
        return {"content": self.visit(node["content"])}

    def _visit_horizontal_rule(self, node):
        return {}

    def _visit_style(self, node):
        return {"node_types": [node["value"]], "content": self.visit(node["content"])}

    def _visit_verbatim(self, node):
        return {"content": node["value"]}

    def _visit_class(self, node):
        return {
            "classes": " ".join(node["classes"]),
            "content": self.visit(node["content"]),
        }

    def _visit_link(self, node):
        return {"text": node["text"], "target": node["target"]}

    def _visit_header(self, node):
        return {
            "level": node["level"],
            "value": node["value"],
            "anchor": node["anchor"],
        }

    def _visit_quote(self, node):
        return {
            "attribution": node["attribution"],
            "content": "".join(self.visitlist(node["content"])),
        }

    def _visit_raw(self, node):
        return {"content": "\n".join(self.visitlist(node["content"]))}

    def _visit_admonition(self, node):
        return {
            "node_types": [
                f'admonition_{node["class"]}',
                "admonition",
            ],
            "class": node["class"],
            "icon": node["icon"],
            "label": node["label"],
            "content": "".join(self.visitlist(node["content"])),
        }

    def _visit_block(self, node):
        return {
            "type": node["blocktype"],
            "content": "".join(self.visitlist(node["content"])),
            "title": self.visit(node["title"]),
            "kwargs": node["kwargs"],
        }

    def _visit_source(self, node):
        lexer = get_lexer_by_name(node["language"])
        formatter = get_formatter_by_name("html")
        src = "\n".join([i["value"] for i in node["code"]])

        highlighted_src = highlight(src, lexer, formatter)

        highlighted_src_lines = highlighted_src.split("\n")

        callouts = {}
        for linenum, callout in node["callouts"].items():
            callout_name, callout_text = callout
            highlighted_src_lines[linenum] = "{line} {callout}".format(
                line=highlighted_src_lines[linenum],
                callout=self._render("callout", name=callout_name),
            )
            callouts[callout_name] = callout_text

        highlighted_src = "\n".join(highlighted_src_lines)
        callouts_list = [
            (self._render("callout", name=name), text)
            for name, text in callouts.items()
        ]

        return {
            "code": highlighted_src,
            "language": node["language"],
            "callouts": callouts_list,
            "title": self.visit(node["title"]),
        }

    def _visit_document(self, node):
        node_type = "document"
        if node["no_document"] is True:
            node_type = "container"

        return {
            "node_types": [node_type],
            "content": "".join([self.visit(item) for item in node["content"]]),
        }

    def _visit_list_item(self, node):
        return {"content": self.visit(node["content"])}

    def _visit_list(self, node):
        return {
            "ordered": node["ordered"],
            "items": "".join([self.visit(i) for i in node["items"]]),
        }

    def _visit_toc_entry(self, node):
        return {
            "anchor": node["anchor"],
            "text": node["text"],
            "children": "".join([self.visit(i) for i in node["children"]]),
        }

    def visit_toc(self, nodes):
        keys = {"entries": "".join([self.visit(i) for i in nodes])}
        return self._render("toc", **keys)

    def _visit_footnote_ref(self, node):
        return {
            "node_types": ["footnote_ref"],
            "number": node["number"],
            "refanchor": node["refanchor"],
            "defanchor": node["defanchor"],
        }

    def _visit_footnote_def(self, node):
        return {
            "number": node["number"],
            "refanchor": node["refanchor"],
            "defanchor": node["defanchor"],
            "text": "".join([self.visit(i) for i in node["content"]]),
        }

    def visit_footnotes(self, nodes):
        keys = {"entries": "".join([self.visit(i) for i in nodes])}
        return self._render("footnotes", **keys)

    def _visit_command(self, node):
        if node["name"] == "toc":
            return {"content": "".join(self.visit_toc(self.toc))}

        if node["name"] == "footnotes":
            return {"content": "".join(self.visit_footnotes(self.footnotes))}

    def _visit_content_image(self, node):
        return {
            "node_types": ["image"],
            "uri": node["uri"],
            "title": self.visit(node["title"]),
            "alt_text": node["alt_text"],
        }

    def _visit_image(self, node):
        return {
            "node_types": ["inline_image"],
            "uri": node["uri"],
            "alt_text": node["alt_text"],
        }

    def _visit_macro(self, node):
        return {}



import pytest

from mau.parsers.base_parser import ParseError
from mau.parsers.arguments_parser import ArgumentsParser

from tests.helpers import init_parser_factory

init_parser = init_parser_factory(ArgumentsParser)


def test_named_argument():
    p = init_parser("argument1=value1")
    p.parse()

    assert p.args == []
    assert p.kwargs == {"argument1": "value1"}


def test_multiple_named_arguments():
    p = init_parser("argument1=value1,argument2=value2")
    p.parse()

    assert p.args == []
    assert p.kwargs == {"argument1": "value1", "argument2": "value2"}


def test_multiple_named_arguments_with_spaces():
    p = init_parser("argument1=value1, argument2=value2")
    p.parse()

    assert p.args == []
    assert p.kwargs == {"argument1": "value1", "argument2": "value2"}


def test_multiple_unnamed_arguments():
    p = init_parser("value1,value2")
    p.parse()

    assert p.args == ["value1", "value2"]
    assert p.kwargs == {}


def test_quotes_with_named_arguments():
    p = init_parser('argument="value1,value2"')
    p.parse()

    assert p.args == []
    assert p.kwargs == {"argument": "value1,value2"}


def test_quotes_with_unnamed_arguments():
    p = init_parser('"value1,value2"')
    p.parse()

    assert p.args == ["value1,value2"]
    assert p.kwargs == {}


def test_escaped_quotes():
    p = init_parser(r'"value \"with quotes\""')
    p.parse()

    assert p.args == ['value "with quotes"']
    assert p.kwargs == {}


def test_multiple_unnamed_and_named_arguments():
    p = init_parser("value1,argument2=value2")
    p.parse()

    assert p.args == ["value1"]
    assert p.kwargs == {"argument2": "value2"}


def test_multiple_named_arguments_before_unnamed_ones():
    p = init_parser("argument1=value1,value2")

    with pytest.raises(ParseError):
        p.parse()


def test_raw_argument():
    p = init_parser("value0, argument1=value1", raw=True)
    p.parse()

    assert p.args == ["value0, argument1=value1"]
    assert p.kwargs == {}



from mau.lexers.base_lexer import Text, Literal, WS
from mau.lexers.arguments_lexer import ArgumentsLexer


def test_single_unnamed_argument():
    lex = ArgumentsLexer()

    lex.process("value1")

    assert lex.tokens == [
        Text("value1"),
    ]


def test_single_named_argument():
    lex = ArgumentsLexer()

    lex.process("argument1=value1")

    assert lex.tokens == [
        Text("argument1"),
        Literal("="),
        Text("value1"),
    ]


def test_multiple_unnamed_arguments():
    lex = ArgumentsLexer()

    lex.process("value1, value2")

    assert lex.tokens == [
        Text("value1"),
        Literal(","),
        WS(" "),
        Text("value2"),
    ]


def test_multiple_named_arguments():
    lex = ArgumentsLexer()

    lex.process("argument1=value1, argument2=value2")

    assert lex.tokens == [
        Text("argument1"),
        Literal("="),
        Text("value1"),
        Literal(","),
        WS(" "),
        Text("argument2"),
        Literal("="),
        Text("value2"),
    ]


def test_mixed_arguments():
    lex = ArgumentsLexer()

    lex.process("value1, value2,argument1=value1, argument2=value2")

    assert lex.tokens == [
        Text("value1"),
        Literal(","),
        WS(" "),
        Text("value2"),
        Literal(","),
        Text("argument1"),
        Literal("="),
        Text("value1"),
        Literal(","),
        WS(" "),
        Text("argument2"),
        Literal("="),
        Text("value2"),
    ]


def test_quotes():
    lex = ArgumentsLexer()

    lex.process('argument1="value1,value2"')

    assert lex.tokens == [
        Text("argument1"),
        Literal("="),
        Literal('"'),
        Text("value1"),
        Literal(","),
        Text("value2"),
        Literal('"'),
    ]

    assert [t.position for t in lex.tokens] == [
        (0, 0),
        (0, 9),
        (0, 10),
        (0, 11),
        (0, 17),
        (0, 18),
        (0, 24),
    ]


def test_spaces():
    lex = ArgumentsLexer()

    lex.process("argument1=value1 value2")

    assert lex.tokens == [
        Text("argument1"),
        Literal("="),
        Text("value1"),
        WS(" "),
        Text("value2"),
    ]


def test_escaped_quotes():
    lex = ArgumentsLexer()

    lex.process(r"Argument \"with\" quotes")

    assert lex.tokens == [
        Text("Argument"),
        WS(" "),
        Literal("\\"),
        Literal('"'),
        Text("with"),
        Literal("\\"),
        Literal('"'),
        WS(" "),
        Text("quotes"),
    ]



from urllib.parse import quote
import copy

from mau.lexers.base_lexer import TokenTypes, Token, TokenError
from mau.lexers.main_lexer import MainLexer
from mau.parsers.base_parser import BaseParser, ParseError, parser, analyse
from mau.parsers.text_parser import TextParser
from mau.parsers.arguments_parser import ArgumentsParser, merge_args
from mau.parsers.preprocess_variables_parser import PreprocessVariablesParser
from mau.parsers.nodes import (
    HorizontalRuleNode,
    TextNode,
    BlockNode,
    SourceNode,
    RawNode,
    AdmonitionNode,
    QuoteNode,
    ContentImageNode,
    CommandNode,
    HeaderNode,
    ListNode,
    ListItemNode,
    ParagraphNode,
    TocNode,
)


def header_anchor(text, level):
    return "h{}-{}-{}".format(
        level, quote(text.lower())[:20], str(id(text))[:8]
    )  # pragma: no cover


class MainParser(BaseParser):
    def __init__(self, variables=None):
        super().__init__()

        self.lexer = MainLexer()

        self.variables = copy.deepcopy(variables) if variables else {}
        self.headers = []
        self.footnotes = []
        self.blocks = {}
        self.toc = []

        self._args = []
        self._kwargs = {}
        self._title = None

    def _add_footnotes(self, footnotes):
        self.footnotes.extend(footnotes)

    def _pop_attributes(self):
        _args = copy.deepcopy(self._args)
        _kwargs = copy.deepcopy(self._kwargs)

        self._args = []
        self._kwargs = {}

        return _args, _kwargs

    def _push_attributes(self, args, kwargs):
        self._args = copy.deepcopy(args)
        self._kwargs = copy.deepcopy(kwargs)

    def _pop_title(self):
        title = self._title
        self._title = None
        return title

    def _push_title(self, title):
        self._title = title

    def collect_lines(self, stop_tokens):
        # This collects several lines of text in a list
        # until it gets to a line that begins with one
        # of the tokens listed in stop_tokens.
        lines = []
        while self.peek_token() not in stop_tokens:
            lines.append(self.collect_join([Token(TokenTypes.EOL)]))
            self.get_token(TokenTypes.EOL)

        return lines

    @parser
    def _parse_horizontal_rule(self):
        self.get_token(TokenTypes.LITERAL, "---")
        self.get_token(TokenTypes.EOL)

        self._save(HorizontalRuleNode())

    @parser
    def _parse_single_line_comment(self):
        self.get_token(TokenTypes.TEXT, check=lambda x: x.startswith("//"))
        self.get_token(TokenTypes.EOL)

    @parser
    def _parse_multi_line_comment(self):
        self.get_token(TokenTypes.LITERAL, "////")
        self.collect_lines([Token(TokenTypes.LITERAL, "////"), Token(TokenTypes.EOF)])
        self.force_token(TokenTypes.LITERAL, "////")

    @parser
    def _parse_title(self):
        self.get_token(TokenTypes.LITERAL, ".")

        with self:
            self.get_token(TokenTypes.WHITESPACE)

        text = self.get_token(TokenTypes.TEXT).value
        self.get_token(TokenTypes.EOL)

        p = analyse(TextParser(footnotes_start_with=len(self.footnotes) + 1), text)
        title = p.nodes[0]

        self._push_title(title)

    @parser
    def _parse_header(self):
        header = self.get_token(
            TokenTypes.LITERAL, check=lambda x: x.startswith("=")
        ).value
        self.get_token(TokenTypes.WHITESPACE)

        in_toc = True
        if header.endswith("!"):
            header = header[:-1]
            in_toc = False

        text = self.get_token(TokenTypes.TEXT).value
        level = len(header)
        anchor = header_anchor(text, level)

        if in_toc:
            self.headers.append((text, level, anchor))

        _, kwargs = self._pop_attributes()

        self._save(HeaderNode(value=text, level=level, anchor=anchor, kwargs=kwargs))

    @parser
    def _parse_content(self):
        self.get_token(TokenTypes.LITERAL, check=lambda x: x.startswith("<<"))
        self.get_token(TokenTypes.WHITESPACE)
        content_type_and_uri = self.get_token(TokenTypes.TEXT).value

        content_type, uri = content_type_and_uri.split(":", maxsplit=1)

        _, kwargs = self._pop_attributes()

        alt_text = kwargs.pop("alt_text", None)
        classes = kwargs.pop("classes", None)
        title = self._pop_title()

        if classes:
            classes = classes.split(",")

        self._save(
            ContentImageNode(
                uri=uri,
                alt_text=alt_text,
                classes=classes,
                title=title,
                kwargs=kwargs,
            )
        )

    @parser
    def _parse_block(self):
        delimiter = self.get_token(TokenTypes.TEXT).value

        if len(delimiter) != 4 or len(set(delimiter)) != 1:
            raise TokenError

        self.get_token(TokenTypes.EOL)

        content = self.collect_lines(
            [Token(TokenTypes.TEXT, delimiter), Token(TokenTypes.EOF)]
        )

        self.force_token(TokenTypes.TEXT, delimiter)
        self.get_token(TokenTypes.EOL)

        secondary_content = self.collect_lines(
            [Token(TokenTypes.EOL), Token(TokenTypes.EOF)]
        )

        args, kwargs = self._pop_attributes()
        title = self._pop_title()

        if len(args) != 0 and args[0] in ["if", "ifnot"]:
            return self._parse_conditional_block(args[0], content, args[1:], kwargs)

        if len(args) != 0 and args[0] in ["raw"]:
            return self._parse_raw_block(content, args[1:], kwargs)

        if len(args) != 0 and args[0] == "source":
            return self._parse_source_block(
                content, secondary_content, title, args[1:], kwargs
            )

        if len(args) != 0 and args[0] == "admonition":
            return self._parse_admonition_block(content, args[1:], kwargs)

        if len(args) != 0 and args[0] == "quote":
            return self._parse_quote_block(content, title, args[1:], kwargs)

        try:
            blocktype = args[0]
            args = args[1:]
        except IndexError:
            blocktype = None

        return self._parse_standard_block(
            blocktype, content, secondary_content, title, args, kwargs
        )

    def _parse_conditional_block(self, condition, content, args, kwargs):
        _, kwargs = merge_args(args, kwargs, ["variable", "value"])

        match = self.variables.get(kwargs["variable"]) == kwargs.get("value", True)
        test = True if condition == "if" else False

        if match is test:
            p = analyse(MainParser(variables=self.variables), "\n".join(content))

            self._add_footnotes(p.footnotes)

            self.nodes.extend(p.nodes)

    def _parse_raw_block(self, content, args, kwargs):
        textlines = [TextNode(line) for line in content]

        self._save(RawNode(content=textlines))

    def _parse_source_block(self, content, secondary_content, title, args, kwargs):
        delimiter = kwargs.pop("callouts", ":")

        # This removes callouts from the source code
        # and maps line numbers to callouts and text
        # {linenum:(callout,text)}
        marked_lines = {}

        for linenum, line in enumerate(content):
            if not line.endswith(delimiter):
                continue

            # Remove the final delimiter
            line = line[:-1]

            splits = line.split(delimiter)
            if len(splits) < 2:
                # It's a trap! There are no separators left
                continue

            callout_name = splits[-1]
            line = delimiter.join(splits[:-1])

            marked_lines[linenum] = (callout_name, line)

        # This maps callouts names to line numbers
        # {callout_name:linenum}
        callout_name_to_linenum = {}
        for linenum, mark in marked_lines.items():
            callout_name, fixed_line = mark
            content[linenum] = fixed_line
            callout_name_to_linenum[callout_name] = linenum

        # This reads the callout text and connects
        # it with the text line
        callouts = {}
        for line in secondary_content:
            if ":" not in line:
                raise ParseError

            callout_name, text = line.split(": ")

            try:
                linenum = callout_name_to_linenum[callout_name]
            except KeyError:
                raise ParseError(f"Callout {callout_name} can't be found")

            callouts[linenum] = (callout_name, text)

        # Source blocks must preserve the content literally
        textlines = [TextNode(line) for line in content]

        _, kwargs = merge_args(args, kwargs, ["language"])

        language = kwargs.pop("language", "text")

        self._save(
            SourceNode(
                language,
                callouts=callouts,
                delimiter=delimiter,
                code=textlines,
                title=title,
                kwargs=kwargs,
            )
        )

    def _parse_admonition_block(self, content, args, kwargs):
        _, kwargs = merge_args(args, kwargs, ["class", "icon", "label"])

        p = analyse(MainParser(variables=self.variables), "\n".join(content))

        self._add_footnotes(p.footnotes)

        self._save(
            AdmonitionNode(
                admclass=kwargs.pop("class"),
                icon=kwargs.pop("icon"),
                label=kwargs.pop("label"),
                content=p.nodes,
                kwargs=kwargs,
            )
        )

    def _parse_quote_block(self, content, title, args, kwargs):
        _, kwargs = merge_args(args, kwargs, ["attribution"])

        p = analyse(MainParser(), "\n".join(content))

        self._save(
            QuoteNode(
                kwargs.pop("attribution"),
                content=p.nodes,
                kwargs=kwargs,
            )
        )

    def _parse_standard_block(
        self, blocktype, content, secondary_content, title, args, kwargs
    ):
        pc = analyse(MainParser(variables=self.variables), "\n".join(content))
        ps = analyse(MainParser(variables=self.variables), "\n".join(secondary_content))

        self._add_footnotes(pc.footnotes)

        self._save(
            BlockNode(
                blocktype=blocktype,
                content=pc.nodes,
                secondary_content=ps.nodes,
                args=args,
                kwargs=kwargs,
                title=title,
            )
        )

    @parser
    def _parse_attributes(self):
        self.get_token(TokenTypes.LITERAL, "[")
        attributes = self.get_token(TokenTypes.TEXT).value
        self.get_token(TokenTypes.LITERAL, "]")

        p = analyse(
            PreprocessVariablesParser(self.variables),
            attributes,
        )

        attributes = p.nodes[0].value

        p = analyse(ArgumentsParser(), attributes)

        self._push_attributes(p.args, p.kwargs)

    def _collect_text_content(self):
        if not self.peek_token_is(TokenTypes.TEXT):
            return None

        values = []

        while self.peek_token_is(TokenTypes.TEXT):
            values.append(self.get_token().value)
            self.get_token(TokenTypes.EOL)

        if len(values) == 0:
            return None

        return " ".join(values)

    def _parse_text_content(self, text):
        if text is None:
            return None

        p = analyse(
            PreprocessVariablesParser(self.variables),
            text,
        )

        text = p.nodes[0].value

        p = analyse(TextParser(footnotes_start_with=len(self.footnotes) + 1), text)

        # Text should return a single sentence node
        result = p.nodes[0]

        if len(p.footnotes) > 0:
            self._add_footnotes(p.footnotes)

        return result

    @parser
    def _parse_command(self):
        self.get_token(TokenTypes.LITERAL, "::")
        name = self.get_token(TokenTypes.TEXT).value
        self.get_token(TokenTypes.LITERAL, ":")

        args = None
        kwargs = None

        with self:
            arguments = self.get_token(TokenTypes.TEXT).value
            p = analyse(ArgumentsParser(), arguments)
            args = p.args
            kwargs = p.kwargs

        self._save(CommandNode(name=name, args=args, kwargs=kwargs))

    @parser
    def _parse_variable_definition(self):
        self.get_token(TokenTypes.LITERAL, ":")
        variable_name = self.get_token(TokenTypes.TEXT).value
        self.get_token(TokenTypes.LITERAL, ":")

        variable_value = True
        if variable_name.startswith("!"):
            variable_value = False
            variable_name = variable_name[1:]

        value = self.collect_join([Token(TokenTypes.EOL)])
        if len(value) > 0:
            variable_value = value

        if "." not in variable_name:
            self.variables[variable_name] = variable_value
        else:
            namespace, variable_name = variable_name.split(".")

            try:
                self.variables[namespace][variable_name] = variable_value
            except KeyError:
                self.variables[namespace] = {variable_name: variable_value}

    def _parse_list_nodes(self):
        with self:
            self.get_token(TokenTypes.WHITESPACE)

        header = self.get_token(TokenTypes.LITERAL, check=lambda x: x[0] in "*#").value
        self.get_token(TokenTypes.WHITESPACE)
        text = self._collect_text_content()

        content = self._parse_text_content(text)
        level = len(header)

        nodes = []
        nodes.append(ListItemNode(level, content))

        while not self.peek_token() in [Token(TokenTypes.EOF), Token(TokenTypes.EOL)]:
            # This is the SentenceNode inside the last node added to the list
            last_node_sentence = nodes[-1].content

            with self:
                self.get_token(TokenTypes.WHITESPACE)

            if len(self.peek_token().value) == level:
                header = self.get_token().value
                self.get_token(TokenTypes.WHITESPACE)
                text = self._collect_text_content()
                content = self._parse_text_content(text)
                nodes.append(ListItemNode(len(header), content))
            elif len(self.peek_token().value) > level:
                numbered = True if self.peek_token().value[0] == "#" else False
                subnodes = self._parse_list_nodes()
                last_node_sentence.content.append(ListNode(numbered, subnodes))
            else:
                break

        return nodes

    @parser
    def _parse_list(self):
        with self:
            self.get_token(TokenTypes.WHITESPACE)

        header = self.peek_token(TokenTypes.LITERAL, check=lambda x: x[0] in "*#")

        numbered = True if header.value[0] == "#" else False
        nodes = self._parse_list_nodes()
        self._save(ListNode(numbered, nodes, main_node=True))

    @parser
    def _parse_paragraph(self):
        lines = self.collect_lines([Token(TokenTypes.EOL), Token(TokenTypes.EOF)])
        text = " ".join(lines)
        sentence = self._parse_text_content(text)

        args, kwargs = self._pop_attributes()

        self._save(ParagraphNode(sentence, args=args, kwargs=kwargs))

    @parser
    def _parse_eol(self):
        self.get_token(TokenTypes.EOL)

    def _parse_functions(self):
        return [
            self._parse_eol,
            self._parse_horizontal_rule,
            self._parse_single_line_comment,
            self._parse_multi_line_comment,
            self._parse_variable_definition,
            self._parse_command,
            self._parse_title,
            self._parse_attributes,
            self._parse_header,
            self._parse_block,
            self._parse_content,
            self._parse_list,
            self._parse_paragraph,
        ]

    def _create_toc(self):
        nodes = []
        latest_by_level = {}

        for i in self.headers:
            text, level, anchor = i

            # This is the current node
            node = TocNode(level, text, anchor)

            # This collects the latest node added with a given level
            latest_by_level[level] = node

            try:
                # Simplest case, add it to the latest one
                # with a level just 1 step lower
                latest_by_level[level - 1].children.append(node)
            except KeyError:
                # Find all the latest ones added with a level lower than this
                latest = [latest_by_level.get(i, None) for i in range(1, level)]

                # Get the children list of each one, plus nodes for the root
                children = [nodes] + [i.children for i in latest if i is not None]

                # Get the nearest one and append to that
                children[-1].append(node)

        return nodes

    def parse(self):
        self._parse()

        self.toc = self._create_toc()



import re
import string
from functools import partial
from collections.abc import Sequence

from mau import text_buffer


class TokenTypes:
    EOL = "EOL"
    EOF = "EOF"
    LITERAL = "LITERAL"
    TEXT = "TEXT"
    WHITESPACE = "WHITESPACE"


class LexerError(ValueError):
    pass


class TokenError(ValueError):
    pass


class Token:
    def __init__(self, _type, value=None, position=None):
        self.type = _type
        self.value = str(value) if value is not None else None
        self.position = position

    def __str__(self):
        position_string = ""
        if self.position:
            position_string = f", line={self.position[0]}, col={self.position[1]}"

        value_string = ""
        if self.value is not None:
            value_string = f", '{self.value}'"

        return f"Token({self.type}{value_string}{position_string})"

    __repr__ = __str__

    def __eq__(self, other):
        if other.value is None:
            return self.type == other.type

        return (self.type, self.value) == (
            other.type,
            other.value,
        )

    def __hash__(self):
        return hash((self.type, self.value))

    def __len__(self):
        if self.value:
            return len(self.value)

        return 0

    def __bool__(self):
        return True


EOL = Token(TokenTypes.EOL)
EOF = Token(TokenTypes.EOF)
WS = partial(Token, TokenTypes.WHITESPACE)
Text = partial(Token, TokenTypes.TEXT)
Literal = partial(Token, TokenTypes.LITERAL)


class BaseLexer:
    def __init__(self, initial_position=None):
        self._text_buffer = text_buffer.TextBuffer()

        self._text_buffer.position
        self._buffer = []
        self._initial_position = initial_position or (0, 0)

        self.tokens = []

    @property
    def _token_position(self):
        return tuple(map(sum, zip(self._text_buffer.position, self._initial_position)))

    def process(self, text):
        self.tokens = []
        self.index = -1
        self._text_buffer.load(text)

        self._process()
        while self.tokens[-1].type is not TokenTypes.EOF:
            self._process()

    def _process(self):
        # This should not be touched by child classes
        # as it is the core of the lexer. It tries
        # each function in the list returned by
        # _process_functions and stores all the resulting
        # tokens. A parsing function must return None
        # when characters do not match the rules.

        process_functions = self._process_functions()
        process_functions.append(self._process_error)

        for process_func in process_functions:
            result = self._wrap(process_func())

            if result is None:
                continue

            self.tokens.extend(result)
            return

    def _wrap(self, result):
        # Makes sure the result is either None or a list of tokens
        if result is None:
            return

        if not isinstance(result, Sequence):
            return [result]

        return result

    def _nextline(self):
        # Carriage return =)
        self._initial_position = (self._initial_position[0], 0)

        # Skip the whole line including the EOL
        self._text_buffer.nextline()

    def _skip(self, steps=1):
        # Skip only the given amount of characters
        self._text_buffer.skip(steps)

    @property
    def _current_char(self):
        return self._text_buffer.current_char

    @property
    def _current_line(self):
        return self._text_buffer.current_line

    @property
    def _tail(self):
        return self._text_buffer.tail

    def _create_token(self, token_type, token_value=None):
        return Token(token_type, token_value, position=self._token_position)

    def _create_token_and_skip(self, token_type, token_value=None, skip_value=None):
        skip = next(x for x in [skip_value, token_value, ""] if x is not None)
        token = self._create_token(token_type, token_value)

        if token_type == TokenTypes.EOL:
            self._nextline()
        else:
            self._skip(len(skip))

        return token

    def _store(self, token_type, token_value=None, skip_value=None):
        self._buffer.append(
            self._create_token_and_skip(token_type, token_value, skip_value)
        )

    def _pop(self):
        tokens = list(self._buffer)
        self._buffer = []
        return tokens

    def _process_eof(self):
        try:
            self._current_line
        except text_buffer.EOFError:
            return self._create_token_and_skip(TokenTypes.EOF)

    def _process_eol(self):
        try:
            self._current_char
        except text_buffer.EOLError:
            return self._create_token_and_skip(TokenTypes.EOL)

    def _process_character(self):
        if self._current_char not in string.ascii_letters:
            return None

        self._store(TokenTypes.TEXT, self._current_char)
        return self._pop()

    def _process_whitespace(self):
        regexp = re.compile(r"\ +")

        match = regexp.match(self._tail)

        if not match:
            return None

        self._store(TokenTypes.WHITESPACE, match.group())
        return self._pop()

    def _process_functions(self):
        return [
            self._process_eof,
            self._process_eol,
            self._process_whitespace,
            self._process_character,
        ]

    def _process_error(self):
        raise LexerError(f'Can\'t process "{self._tail}"')



# -*- coding: utf-8 -*-
from pkg_resources import DistributionNotFound, get_distribution

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = get_distribution(dist_name).version
except DistributionNotFound:
    __version__ = "unknown"
finally:
    del get_distribution, DistributionNotFound

from mau.parsers.main_parser import MainParser
from mau.parsers.nodes import DocumentNode
from mau.visitors.html_visitor import HTMLVisitor
from mau.visitors.asciidoctor_visitor import AsciidoctorVisitor


class Mau:
    def __init__(
        self, config, target_format, default_templates=None, templates_directory=None
    ):
        self.config = config
        self.target_format = target_format
        self.default_templates = default_templates
        self.templates_directory = templates_directory

        self.no_document = config.get("no_document", False)

        self.variables = {}

    def process(self, text):
        self.config["mau"] = {"target_format": self.target_format}

        parser = MainParser(variables=self.config)
        parser.load(text)
        parser.parse()

        self.variables = parser.variables

        document = DocumentNode(parser.nodes, self.no_document)

        ast = document.asdict()

        if self.target_format == "html":
            visitor_class = HTMLVisitor
        elif self.target_format == "asciidoctor":
            visitor_class = AsciidoctorVisitor

        visitor = visitor_class(
            default_templates=self.default_templates,
            templates_directory=self.templates_directory,
            config=self.config,
            toc=[i.asdict() for i in parser.toc],
            footnotes=[i.asdict() for i in parser.footnotes],
        )

        return visitor.visit(ast)



import re

from mau.lexers.base_lexer import BaseLexer, TokenTypes


class ArgumentsLexer(BaseLexer):
    def _process_functions(self):
        return [
            self._process_eof,
            self._process_eol,
            self._process_whitespace,
            self._process_literal,
            self._process_text,
        ]

    def _process_whitespace(self):
        regexp = re.compile(r" +")

        match = regexp.match(self._tail)

        if not match:  # pragma: no cover
            return None

        return self._create_token_and_skip(TokenTypes.WHITESPACE, match.group())

    def _process_literal(self):
        if self._current_char not in r'\=,"':
            return None

        return self._create_token_and_skip(TokenTypes.LITERAL, self._current_char)

    def _process_text(self):
        regexp = re.compile(r'[^\\=," ]+')

        match = regexp.match(self._tail)

        if not match:  # pragma: no cover
            return None

        return self._create_token_and_skip(TokenTypes.TEXT, match.group())

    def process(self, text):
        super().process(text)

        # Remove EOL and EOF
        self.tokens = self.tokens[:-2]



import pytest

from mau.parsers.base_parser import ParseError
from mau.parsers.main_parser import MainParser

from tests.helpers import listasdict, init_parser_factory, parser_test_factory, dedent

init_parser = init_parser_factory(MainParser)

_test = parser_test_factory(MainParser)


def test_parse_discards_empty_lines():
    _test("\n\n\n\n", [])


def test_horizontal_rule():
    _test("---", [{"type": "horizontal_rule"}])


def test_parse_single_line_comments():
    _test("// Just a comment", [])


def test_parse_multi_line_comments():
    p = init_parser(
        dedent(
            """
            ////
            This is a
            multiline
            comment
            ////
            """
        )
    )
    p.parse()

    assert listasdict(p.nodes) == []


def test_parse_multi_line_comments_with_dangerous_characters():
    source = """
    ////
    .This is a
    [multiline]
    ----
    comment
    ////
    """

    _test(source, [])


def test_parse_open_multi_line_comments():
    p = init_parser(
        dedent(
            """
            ////
            This is a
            multiline
            comment
            """
        )
    )

    with pytest.raises(ParseError):
        p.parse()


def test_parse_paragraphs_starting_with_a_macro():
    p = init_parser("[link](http://some.where,text) is the link I want")
    p.parse()

    assert listasdict(p.nodes) == [
        {
            "type": "paragraph",
            "args": [],
            "kwargs": {},
            "content": {
                "type": "sentence",
                "content": [
                    {
                        "type": "link",
                        "text": "text",
                        "target": "http://some.where",
                    },
                    {
                        "type": "text",
                        "value": " is the link I want",
                    },
                ],
            },
        },
    ]


def test_parse_paragraphs_is_a_macro_only():
    p = init_parser("[link](http://some.where,text)")
    p.parse()

    assert listasdict(p.nodes) == [
        {
            "type": "paragraph",
            "args": [],
            "kwargs": {},
            "content": {
                "type": "sentence",
                "content": [
                    {
                        "type": "link",
                        "text": "text",
                        "target": "http://some.where",
                    },
                ],
            },
        },
    ]


def test_command():
    p = init_parser("::somecommand:argument1=value1,argument2=value2")
    p.parse()

    assert listasdict(p.nodes) == [
        {
            "type": "command",
            "name": "somecommand",
            "kwargs": {"argument1": "value1", "argument2": "value2"},
        },
    ]


def test_command_without_arguments():
    p = init_parser("::somecommand:")
    p.parse()

    assert listasdict(p.nodes) == [
        {
            "type": "command",
            "name": "somecommand",
            "kwargs": {},
        },
    ]


def test_attributes_paragraph():
    p = init_parser(
        dedent(
            """
            [value1,someattr1=somevalue1,someattr2=somevalue2]
            This is text
            """
        )
    )
    p.parse()

    assert listasdict(p.nodes) == [
        {
            "type": "paragraph",
            "args": ["value1"],
            "kwargs": {"someattr1": "somevalue1", "someattr2": "somevalue2"},
            "content": {
                "type": "sentence",
                "content": [
                    {
                        "type": "text",
                        "value": "This is text",
                    }
                ],
            },
        },
    ]


def test_attributes_with_variables():
    p = init_parser(
        dedent(
            """
            :attrs:value1,someattr1=somevalue1,someattr2=somevalue2

            [{attrs}]
            This is text
            """
        )
    )
    p.parse()

    assert listasdict(p.nodes) == [
        {
            "type": "paragraph",
            "args": ["value1"],
            "kwargs": {"someattr1": "somevalue1", "someattr2": "somevalue2"},
            "content": {
                "type": "sentence",
                "content": [
                    {
                        "type": "text",
                        "value": "This is text",
                    }
                ],
            },
        },
    ]


def test_simple_paragraph():
    source = """
    This is a simple line of text
    """

    expected = [
        {
            "type": "paragraph",
            "args": [],
            "kwargs": {},
            "content": {
                "type": "sentence",
                "content": [
                    {
                        "type": "text",
                        "value": "This is a simple line of text",
                    }
                ],
            },
        }
    ]

    _test(source, expected)


def test_multiline_paragraph():
    source = """
    This is a simple line of text
    followed by another line of text
    """

    expected = [
        {
            "type": "paragraph",
            "args": [],
            "kwargs": {},
            "content": {
                "type": "sentence",
                "content": [
                    {
                        "type": "text",
                        "value": "This is a simple line of text followed by another line of text",
                    }
                ],
            },
        }
    ]

    _test(source, expected)


def test_multiple_paragraphs():
    source = """
    This is a simple line of text
    followed by another line of text

    And this is another paragraph
    """

    expected = [
        {
            "type": "paragraph",
            "args": [],
            "kwargs": {},
            "content": {
                "type": "sentence",
                "content": [
                    {
                        "type": "text",
                        "value": "This is a simple line of text followed by another line of text",
                    }
                ],
            },
        },
        {
            "type": "paragraph",
            "args": [],
            "kwargs": {},
            "content": {
                "type": "sentence",
                "content": [
                    {
                        "type": "text",
                        "value": "And this is another paragraph",
                    }
                ],
            },
        },
    ]

    _test(source, expected)


def test_style_underscore():
    source = """
    This is _underscore_ text
    """

    expected = [
        {
            "type": "paragraph",
            "args": [],
            "kwargs": {},
            "content": {
                "type": "sentence",
                "content": [
                    {
                        "type": "text",
                        "value": "This is ",
                    },
                    {
                        "type": "style",
                        "value": "underscore",
                        "content": {
                            "type": "sentence",
                            "content": [{"type": "text", "value": "underscore"}],
                        },
                    },
                    {
                        "type": "text",
                        "value": " text",
                    },
                ],
            },
        }
    ]

    _test(source, expected)


def test_style_star():
    source = """
    This is *star* text
    """

    expected = [
        {
            "type": "paragraph",
            "args": [],
            "kwargs": {},
            "content": {
                "type": "sentence",
                "content": [
                    {
                        "type": "text",
                        "value": "This is ",
                    },
                    {
                        "type": "style",
                        "value": "star",
                        "content": {
                            "type": "sentence",
                            "content": [{"type": "text", "value": "star"}],
                        },
                    },
                    {
                        "type": "text",
                        "value": " text",
                    },
                ],
            },
        }
    ]

    _test(source, expected)


def test_style_at_beginning():
    source = """
    *This is star text*
    """

    expected = [
        {
            "type": "paragraph",
            "args": [],
            "kwargs": {},
            "content": {
                "type": "sentence",
                "content": [
                    {
                        "type": "style",
                        "value": "star",
                        "content": {
                            "type": "sentence",
                            "content": [{"type": "text", "value": "This is star text"}],
                        },
                    },
                ],
            },
        }
    ]

    _test(source, expected)


def test_style_not_closed():
    source = r"""
    This ` is a backtick and this _an underscore
    """

    expected = [
        {
            "type": "paragraph",
            "args": [],
            "kwargs": {},
            "content": {
                "type": "sentence",
                "content": [
                    {
                        "type": "text",
                        "value": "This ` is a backtick and this _an underscore",
                    },
                ],
            },
        }
    ]

    _test(source, expected)


def test_style_escape_markers():
    source = r"""
    This is \_underscore\_ and this is \`verbatim\`
    """

    expected = [
        {
            "type": "paragraph",
            "args": [],
            "kwargs": {},
            "content": {
                "type": "sentence",
                "content": [
                    {
                        "type": "text",
                        "value": "This is _underscore_ and this is `verbatim`",
                    },
                ],
            },
        }
    ]

    _test(source, expected)



from mau.visitors.visitor import Visitor

DEFAULT_TEMPLATES = {
    "admonition.adoc": "[{{ admclass }}{% if icon %}.{{ icon }}{% endif %}]\n====\n{{ content|join }}====\n",
    "block.adoc": "--\n{{ content|join('\n') }}\n--",
    "callout.adoc": "<{{ name }}>",
    "class.adoc": """[{{ classes|join(' ') }}]#{{ content }}#""",
    "document.adoc": "{{ content|join('\n') }}",
    "footnote_ref.adoc": "footnote:[{{ text|join }}]",
    "header.adoc": "{{ header }} {{ text }}\n",
    "horizontal_rule.adoc": "---\n",
    "image.adoc": "{% if asciidoctor_classes %}[{{ asciidoctor_classes }}]\n{% endif %}{% if title %}.{{ title }}\n{% endif %}image::{{ uri }}[{% if alt_text %}{{ alt_text }}{% endif %}]\n",
    "inline_image.adoc": "image::{{ uri }}{%if alt_text %}[{{ alt_text }}]{% endif %}",
    "link.adoc": "{{ target }}{% if text %}[{{ text }}]{% endif %}",
    "list.adoc": "{{ items|join }}{% if main_node %}\n{% endif %}",
    "list_item.adoc": "{% if not main_node %}\n{% endif %}{% if prefix %}{{ prefix }} {% endif %}{{ content }}",
    "paragraph.adoc": "{{ content }}\n",
    "quote.adoc": '[quote, "{{ attribution }}"]\n____\n{{ content|join }}____\n',
    "sentence.adoc": "{{ content }}",
    "source.adoc": "{% if title %}.{{ title }}\n{% endif %}[source{% if language %},{{ language }}{% endif %}]\n----\n{{ code }}\n----\n{% if callouts %}{% for callout in callouts %}{{ callout[0] }} {{ callout[1] }}{% endfor %}\n{% endif %}",
    "star.adoc": "*{{ content }}*",
    "text.adoc": "{{ value }}",
    "underscore.adoc": "_{{ content }}_",
    "verbatim.adoc": "`{{ content }}`",
}


class AsciidoctorVisitor(Visitor):
    _template_extension = ".adoc"
    _environment_options = {"keep_trailing_newline": True}

    def __init__(
        self,
        default_templates=None,
        templates_directory=None,
        config=None,
        toc=None,
        footnotes=None,
    ):
        super().__init__(
            default_templates=DEFAULT_TEMPLATES,
            templates_directory=templates_directory,
            config=config,
            toc=toc,
            footnotes=footnotes,
        )

    def _visit_text(self, node):
        return {"value": node["value"]}

    def _visit_sentence(self, node):
        return {"content": "".join([self.visit(t) for t in node["content"]])}

    def _visit_paragraph(self, node):
        return {"content": self.visit(node["content"])}

    def _visit_horizontal_rule(self, node):
        return {}

    def _visit_style(self, node):
        return {"node_types": [node["value"]], "content": self.visit(node["content"])}

    def _visit_verbatim(self, node):
        return {"content": node["value"]}

    def _visit_class(self, node):
        classes = [f".{cls}" for cls in node["classes"]]
        return {"classes": classes, "content": self.visit(node["content"])}

    def _visit_link(self, node):
        text = node["text"]
        target = node["target"]

        if text == target:
            text = None

        return {"text": text, "target": node["target"]}

    def _visit_header(self, node):
        return {"header": "=" * int(node["level"]), "text": node["value"]}

    def _visit_quote(self, node):
        return {
            "attribution": node["attribution"],
            "content": self.visitlist(node["content"]),
        }

    def _visit_admonition(self, node):
        admclass = node["class"]
        content = self.visitlist(node["content"])

        if admclass in ["note", "tip", "important", "caution", "warning"]:
            admclass = admclass.upper()
        else:
            raise ValueError(f"Admonition {admclass} cannot be converted")

        return {
            "admclass": admclass,
            "icon": node["icon"],
            "content": content,
        }

    def _visit_block(self, node):
        return {
            "content": self.visitlist(node["content"]),
            "title": self.visit(node["title"]),
            "kwargs": node["kwargs"],
        }

    def _visit_source(self, node):
        src = [i["value"] for i in node["code"]]

        callouts = {}
        for linenum, callout in node["callouts"].items():
            callout_name, callout_text = callout
            src[linenum] = "{line} {callout}".format(
                line=src[linenum],
                callout=self._render("callout", name=callout_name),
            )
            callouts[callout_name] = callout_text

        src = "\n".join(src)
        callouts_list = [
            (self._render("callout", name=name), text)
            for name, text in callouts.items()
        ]

        return {
            "code": src,
            "title": self.visit(node["title"]),
            "language": node["language"],
            "callouts": callouts_list,
        }

    def _visit_document(self, node):
        return {"content": [self.visit(item) for item in node["content"]]}

    def _visit_list_item(self, node, ordered=False):
        mark = "*"
        if ordered:
            mark = "."

        prefix = None
        if node["content"]["type"] != "list":
            prefix = mark * int(node["level"])

        return {"content": self.visit(node["content"]), "prefix": prefix}

    def _visit_list(self, node, ordered=False):
        return {
            "items": [self.visit(i, ordered=node["ordered"]) for i in node["items"]],
            "main_node": node["main_node"],
        }

    def _visit_footnote_ref(self, node):
        number = node["number"]
        footnote = [i for i in self.footnotes if i["number"] == number][0]

        return {
            "node_types": ["footnote_ref"],
            "text": [self.visit(i) for i in footnote["content"]],
        }

    def _visit_content_image(self, node):
        return {
            "node_types": ["image"],
            "uri": node["uri"],
            "title": self.visit(node["title"]),
            "asciidoctor_classes": node["kwargs"].get("asciidoctor_classes", None),
            "alt_text": node["alt_text"],
        }

    def _visit_image(self, node):
        return {
            "node_types": ["inline_image"],
            "uri": node["uri"],
            "alt_text": node["alt_text"],
        }



from mau.lexers.base_lexer import Token, TokenTypes
from mau.lexers.preprocess_variables_lexer import PreprocessVariablesLexer
from mau.parsers.base_parser import BaseParser, parser
from mau.parsers.nodes import TextNode


class PreprocessError(ValueError):
    pass


class PreprocessVariablesParser(BaseParser):
    def __init__(self, variables=None):
        super().__init__()

        self.lexer = PreprocessVariablesLexer()

        self._text = []
        self.variables = variables or {}

    @parser
    def _parse_verbatim(self):
        self.get_token(TokenTypes.LITERAL, "`")
        text = self.collect_join(
            [Token(TokenTypes.LITERAL, "`"), Token(TokenTypes.EOF)],
            preserve_escaped_stop_tokens=True,
        )
        self.get_token(TokenTypes.LITERAL, "`")

        text = f"`{text}`"

        self._save(TextNode(text))

    @parser
    def _parse_escaped_char(self):
        self.get_token(TokenTypes.LITERAL, "\\")

        char = self.get_token().value

        if char not in "{}":
            char = f"\\{char}"

        self._save(TextNode(char))

    @parser
    def _parse_curly(self):
        variable_name = []
        self.get_token(TokenTypes.LITERAL, "{")
        variable_name = self.collect_join(
            [Token(TokenTypes.LITERAL, "}"), Token(TokenTypes.EOF)]
        )
        self.get_token(TokenTypes.LITERAL, "}")

        try:
            if "." not in variable_name:
                variable_value = self.variables[variable_name]
            else:
                namespace, variable_name = variable_name.split(".")

                variable_value = self.variables[namespace][variable_name]

            self._save(TextNode(variable_value))
        except KeyError:
            raise PreprocessError(f'Attribute "{variable_name}" has not been defined')

    @parser
    def _parse_pass(self):
        self._save(TextNode(self.get_token().value))

    def _parse_functions(self):
        return [
            self._parse_escaped_char,
            self._parse_verbatim,
            self._parse_curly,
            self._parse_pass,
        ]

    def parse(self):
        self._parse()
        text = "".join([str(i.value) for i in self.nodes])
        self.nodes = [TextNode(text)]



from unittest.mock import patch

from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_test_factory

init_parser = init_parser_factory(MainParser)

_test = parser_test_factory(MainParser)


@patch("mau.parsers.main_parser.header_anchor")
def test_parse_header_level_1(header_anchor_mock):
    header_anchor_mock.return_value = "XXXXXX"

    source = """
    = Title of the section
    """

    expected = [
        {
            "type": "header",
            "kwargs": {},
            "value": "Title of the section",
            "level": 1,
            "anchor": "XXXXXX",
        }
    ]

    _test(source, expected)


@patch("mau.parsers.main_parser.header_anchor")
def test_parse_header_level_3(header_anchor_mock):
    header_anchor_mock.return_value = "XXXXXX"

    source = """
    === Title of a subsection
    """

    expected = [
        {
            "type": "header",
            "kwargs": {},
            "value": "Title of a subsection",
            "level": 3,
            "anchor": "XXXXXX",
        }
    ]

    _test(source, expected)


@patch("mau.parsers.main_parser.header_anchor")
def test_parse_collect_headers(header_anchor_mock):
    header_anchor_mock.side_effect = lambda text, level: f"{text}-XXXXXX"

    source = """
    = Header 1
    == Header 1.1
    == Header 1.2
    = Header 2
    == Header 2.1
    === Header 2.1.1
    """

    expected = [
        {
            "type": "header",
            "kwargs": {},
            "value": "Header 1",
            "level": 1,
            "anchor": "Header 1-XXXXXX",
        },
        {
            "type": "header",
            "kwargs": {},
            "value": "Header 1.1",
            "level": 2,
            "anchor": "Header 1.1-XXXXXX",
        },
        {
            "type": "header",
            "kwargs": {},
            "value": "Header 1.2",
            "level": 2,
            "anchor": "Header 1.2-XXXXXX",
        },
        {
            "type": "header",
            "kwargs": {},
            "value": "Header 2",
            "level": 1,
            "anchor": "Header 2-XXXXXX",
        },
        {
            "type": "header",
            "kwargs": {},
            "value": "Header 2.1",
            "level": 2,
            "anchor": "Header 2.1-XXXXXX",
        },
        {
            "type": "header",
            "kwargs": {},
            "value": "Header 2.1.1",
            "level": 3,
            "anchor": "Header 2.1.1-XXXXXX",
        },
    ]

    p = _test(source, expected)

    assert p.headers == [
        ("Header 1", 1, "Header 1-XXXXXX"),
        ("Header 1.1", 2, "Header 1.1-XXXXXX"),
        ("Header 1.2", 2, "Header 1.2-XXXXXX"),
        ("Header 2", 1, "Header 2-XXXXXX"),
        ("Header 2.1", 2, "Header 2.1-XXXXXX"),
        ("Header 2.1.1", 3, "Header 2.1.1-XXXXXX"),
    ]


@patch("mau.parsers.main_parser.header_anchor")
def test_attributes_header(header_anchor_mock):
    header_anchor_mock.side_effect = lambda text, level: f"{text}-XXXXXX"

    source = """
    [value1,someattr1=somevalue1,someattr2=somevalue2]
    = Header
    """

    expected = [
        {
            "type": "header",
            "value": "Header",
            "level": 1,
            "kwargs": {"someattr1": "somevalue1", "someattr2": "somevalue2"},
            "anchor": "Header-XXXXXX",
        },
    ]

    _test(source, expected)



import textwrap

import pytest

from mau import text_buffer


def dedent(text):
    return textwrap.dedent(text).strip()


def test_text_buffer_init_empty():
    tb = text_buffer.TextBuffer()

    with pytest.raises(text_buffer.EOFError):
        tb.current_char


def test_text_buffer_init_one_line():
    tb = text_buffer.TextBuffer("abcdef")

    assert tb.current_char == "a"
    assert tb.peek_char == "b"
    assert tb.current_line == "abcdef"
    assert tb.line == 0
    assert tb.column == 0


def test_text_buffer_end_of_line_peek_char():
    tb = text_buffer.TextBuffer("abcdef")
    tb.column = 5

    assert tb.current_char == "f"
    with pytest.raises(text_buffer.EOLError):
        tb.peek_char


def test_text_buffer_end_of_line_current_char():
    tb = text_buffer.TextBuffer("abcdef")
    tb.column = 200

    with pytest.raises(text_buffer.EOLError):
        tb.current_char


def test_text_buffer_error_at_end_of_file():
    tb = text_buffer.TextBuffer("abcdef")
    tb.line = 1

    with pytest.raises(text_buffer.EOFError):
        tb.current_line


def test_text_buffer_error_after_end_of_file():
    tb = text_buffer.TextBuffer("abcdef")
    tb.line = 100

    with pytest.raises(text_buffer.EOFError):
        tb.current_line


def test_text_buffer_tail():
    ts = text_buffer.TextBuffer("abcdefgh")
    ts.column = 4

    assert ts.tail == "efgh"


def test_text_buffer_multiple_lines():
    tb = text_buffer.TextBuffer("abc\ndef\nghi")
    tb.line = 1
    tb.column = 1

    assert tb.current_line == "def"
    assert tb.current_char == "e"
    assert tb.peek_char == "f"


def test_text_buffer_nextline():
    tb = text_buffer.TextBuffer("abc\ndef\nghi")
    tb.line = 1
    tb.column = 2
    tb.nextline()

    assert tb.position == (2, 0)


def test_text_buffer_position():
    tb = text_buffer.TextBuffer()
    tb.line = 123
    tb.column = 456

    assert tb.position == (123, 456)


def test_text_buffer_set_position():
    tb = text_buffer.TextBuffer()
    tb.position = (222, 333)

    assert tb.position == (222, 333)


def test_text_buffer_skip_defaults_to_one():
    tb = text_buffer.TextBuffer("abc\ndef\nghi")
    tb.skip()

    assert tb.column == 1


def test_text_buffer_skip_accepts_value():
    tb = text_buffer.TextBuffer("abc\ndef\nghi")
    tb.skip(3)

    assert tb.column == 3


def test_text_buffer_goto():
    tb = text_buffer.TextBuffer()
    tb.goto(12, 45)

    assert tb.position == (12, 45)


def test_text_buffer_goto_default_column():
    tb = text_buffer.TextBuffer()
    tb.goto(12)

    assert tb.position == (12, 0)


def test_context():
    tb = text_buffer.TextBuffer(
        textwrap.dedent(
            """
            Some text
            on multiple lines
            that I will use
            to check the context()
            """
        )
    )

    context = dedent(
        """
        2: on multiple lines
        3: that I will use
                  ^
        """
    )

    assert tb.context(3, 7) == context.split("\n")


def test_context_pair_line_numbers():
    tb = text_buffer.TextBuffer(
        textwrap.dedent(
            """
            Some text
            on multiple lines
            that I will use
            to check the context()
            and to text that
            the whole thing
            works with bigger
            line numbers
            in front of the lines
            I need to add
            more text
            """
        )
    )

    context = dedent(
        """
        09: in front of the lines
        10: I need to add
                   ^
        """
    )

    assert tb.context(10, 7) == context.split("\n")


def test_context_long_lines():
    tb = text_buffer.TextBuffer(
        textwrap.dedent(
            """
            Some text on multiple lines that I will use to check the context() and to text that the whole thing
            works with bigger line numbers in front of the lines I need to add more text
            """
        )
    )

    context = dedent(
        """
        1: [...] e lines that I will use to check the cont [...]
        2: [...] ne numbers in front of the lines I need t [...]
                                     ^
        """
    )

    assert tb.context(2, 40) == context.split("\n")


def test_context_long_lines_beginning():
    tb = text_buffer.TextBuffer(
        textwrap.dedent(
            """
            Some text on multiple lines that I will use to check the context() and to text that the whole thing
            works with bigger line numbers in front of the lines I need to add more text
            """
        )
    )

    context = dedent(
        """
        1: Some text on multiple lines that I will [...]
        2: works with bigger line numbers in front [...]
                             ^
        """
    )

    assert tb.context(2, 18) == context.split("\n")


def test_context_long_lines_end():
    tb = text_buffer.TextBuffer(
        textwrap.dedent(
            """
            Some text on multiple lines that I will use to check the context() and to text that the whole thing
            works with bigger line numbers in front of the lines I need to add more text
            """
        )
    )

    context = dedent(
        """
        1: [...] use to check the context() and to text th [...]
        2: [...] of the lines I need to add more text
                                     ^
        """
    )

    assert tb.context(2, 60) == context.split("\n")



from mau.visitors.html_visitor import HTMLVisitor

from mau.parsers.main_parser import MainParser
from tests.helpers import remove_indentation, init_parser_factory, visitlist_factory

init_parser = init_parser_factory(MainParser)
visitlist = visitlist_factory(HTMLVisitor)


def test_footnote():
    v = HTMLVisitor()

    result = v.visit(
        {
            "type": "footnote_ref",
            "number": 6,
            "refanchor": "refXYZ",
            "defanchor": "defXYZ",
        }
    )

    assert result == '<sup>[<a id="refXYZ" href="#defXYZ">6</a>]</sup>'


def test_footnote_definition():
    v = HTMLVisitor()

    result = v._visit_footnote_def(
        {
            "type": "footnote_def",
            "number": 1,
            "refanchor": "refXYZ",
            "defanchor": "defXYZ",
            "content": [
                {
                    "type": "sentence",
                    "content": [
                        {"type": "text", "value": "Some text 1"},
                    ],
                }
            ],
        }
    )

    assert result == {
        "defanchor": "defXYZ",
        "number": 1,
        "refanchor": "refXYZ",
        "text": "Some text 1",
    }


def test_footnotes():
    v = HTMLVisitor()

    result = v.visit_footnotes(
        [
            {
                "type": "footnote_def",
                "number": 1,
                "refanchor": "refXYZ1",
                "defanchor": "defXYZ1",
                "content": [
                    {
                        "type": "sentence",
                        "content": [
                            {"type": "text", "value": "Some text 1"},
                        ],
                    }
                ],
            },
            {
                "type": "footnote_def",
                "number": 2,
                "refanchor": "refXYZ2",
                "defanchor": "defXYZ2",
                "content": [
                    {
                        "type": "sentence",
                        "content": [
                            {"type": "text", "value": "Some text 2"},
                        ],
                    }
                ],
            },
        ]
    )

    assert result == remove_indentation(
        """
        <div id="_footnotes">
          <div id="defXYZ1">
            <a href="#refXYZ1">1</a> Some text 1
          </div>
          <div id="defXYZ2">
            <a href="#refXYZ2">2</a> Some text 2
          </div>
        </div>
        """
    )


def test_command_footnotes():
    footnotes = [
        {
            "type": "footnote_def",
            "number": 1,
            "refanchor": "refXYZ1",
            "defanchor": "defXYZ1",
            "content": [
                {
                    "type": "sentence",
                    "content": [
                        {"type": "text", "value": "Some text 1"},
                    ],
                }
            ],
        },
        {
            "type": "footnote_def",
            "number": 2,
            "refanchor": "refXYZ2",
            "defanchor": "defXYZ2",
            "content": [
                {
                    "type": "sentence",
                    "content": [
                        {"type": "text", "value": "Some text 2"},
                    ],
                }
            ],
        },
    ]

    parser = init_parser("::footnotes:")
    parser.parse()

    result = visitlist([node.asdict() for node in parser.nodes], footnotes=footnotes)

    assert result == [
        remove_indentation(
            """
            <div id="_footnotes">
              <div id="defXYZ1">
                <a href="#refXYZ1">1</a> Some text 1
              </div>
              <div id="defXYZ2">
                <a href="#refXYZ2">2</a> Some text 2
              </div>
            </div>
            """
        )
    ]



import pytest

from mau.parsers.main_parser import MainParser, ParseError

from tests.helpers import init_parser_factory, parser_test_factory

init_parser = init_parser_factory(MainParser)

_test = parser_test_factory(MainParser)


def test_parse_block_with_empty_body():
    source = """
    ----
    ----
    """

    expected = [
        {
            "type": "block",
            "blocktype": None,
            "content": [],
            "secondary_content": [],
            "title": None,
            "args": [],
            "kwargs": {},
        },
    ]

    _test(source, expected)


def test_parse_block_content():
    source = """
    ----
    This is a paragraph.
    ----
    """

    expected = [
        {
            "type": "block",
            "blocktype": None,
            "content": [
                {
                    "type": "paragraph",
                    "args": [],
                    "kwargs": {},
                    "content": {
                        "content": [
                            {
                                "type": "text",
                                "value": "This is a paragraph.",
                            },
                        ],
                        "type": "sentence",
                    },
                },
            ],
            "secondary_content": [],
            "title": None,
            "args": [],
            "kwargs": {},
        },
    ]

    _test(source, expected)


def test_parse_block_content_variables():
    source = """
    ----
    :answer:42
    The answer is {answer}.
    ----
    """

    expected = [
        {
            "type": "block",
            "blocktype": None,
            "content": [
                {
                    "type": "paragraph",
                    "args": [],
                    "kwargs": {},
                    "content": {
                        "content": [
                            {
                                "type": "text",
                                "value": "The answer is 42.",
                            },
                        ],
                        "type": "sentence",
                    },
                },
            ],
            "secondary_content": [],
            "title": None,
            "args": [],
            "kwargs": {},
        },
    ]

    _test(source, expected)


def test_parse_block_content_external_variables():
    source = """
    :answer:42
    ----
    The answer is {answer}.
    ----
    """

    expected = [
        {
            "type": "block",
            "blocktype": None,
            "content": [
                {
                    "type": "paragraph",
                    "args": [],
                    "kwargs": {},
                    "content": {
                        "content": [
                            {
                                "type": "text",
                                "value": "The answer is 42.",
                            },
                        ],
                        "type": "sentence",
                    },
                },
            ],
            "secondary_content": [],
            "title": None,
            "args": [],
            "kwargs": {},
        },
    ]

    _test(source, expected)


def test_block_without_closing_fences():
    p = init_parser("----")

    with pytest.raises(ParseError):
        p.parse()


def test_parse_block_secondary_content():
    source = """
    ----
    ----
    This is a paragraph that gets eaten.
    This is a second paragraph that gets eaten.
    
    This paragraph appears in the output.
    """

    expected = [
        {
            "type": "block",
            "title": None,
            "blocktype": None,
            "args": [],
            "kwargs": {},
            "content": [],
            "secondary_content": [
                {
                    "type": "paragraph",
                    "args": [],
                    "kwargs": {},
                    "content": {
                        "content": [
                            {
                                "type": "text",
                                "value": "This is a paragraph that gets eaten. This is a second paragraph that gets eaten.",
                            },
                        ],
                        "type": "sentence",
                    },
                },
            ],
        },
        {
            "type": "paragraph",
            "args": [],
            "kwargs": {},
            "content": {
                "type": "sentence",
                "content": [
                    {
                        "type": "text",
                        "value": "This paragraph appears in the output.",
                    }
                ],
            },
        },
    ]

    _test(source, expected)


def test_parse_block_inside_block():
    source = """
    ----
    ++++
    ++++
    ----
    """

    expected = [
        {
            "type": "block",
            "title": None,
            "blocktype": None,
            "args": [],
            "kwargs": {},
            "secondary_content": [],
            "content": [
                {
                    "type": "block",
                    "blocktype": None,
                    "title": None,
                    "content": [],
                    "secondary_content": [],
                    "args": [],
                    "kwargs": {},
                },
            ],
        }
    ]

    _test(source, expected)


def test_attributes_block():
    source = """
    [blocktype,myattr1=value1]
    ----
    This is a simple line of text
    followed by another line of text

    And this is another paragraph
    ----
    """

    expected = [
        {
            "type": "block",
            "args": [],
            "blocktype": "blocktype",
            "kwargs": {"myattr1": "value1"},
            "secondary_content": [],
            "title": None,
            "content": [
                {
                    "type": "paragraph",
                    "args": [],
                    "kwargs": {},
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "text",
                                "value": "This is a simple line of text followed by another line of text",
                            }
                        ],
                    },
                },
                {
                    "type": "paragraph",
                    "args": [],
                    "kwargs": {},
                    "content": {
                        "type": "sentence",
                        "content": [
                            {
                                "type": "text",
                                "value": "And this is another paragraph",
                            }
                        ],
                    },
                },
            ],
        },
    ]

    _test(source, expected)



from unittest.mock import patch

from mau.parsers.text_parser import TextParser

from tests.helpers import listasdict, init_parser_factory, parser_test_factory

init_parser = init_parser_factory(TextParser)

_test = parser_test_factory(TextParser)


def test_macro():
    source = "[macroname](value1,value2)"

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "macro",
                    "value": "macroname",
                    "args": ["value1", "value2"],
                    "kwargs": {},
                }
            ],
        }
    ]

    _test(source, expected)


def test_macro_arguments_with_quotes():
    source = '[macroname]("value1,value2")'

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "macro",
                    "value": "macroname",
                    "args": ["value1,value2"],
                    "kwargs": {},
                }
            ],
        }
    ]

    _test(source, expected)


def test_macro_link():
    source = '[link](https://somedomain.org/the/path, "link text")'

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "link",
                    "target": "https://somedomain.org/the/path",
                    "text": "link text",
                }
            ],
        }
    ]

    _test(source, expected)


def test_macro_link_with_quotes():
    source = '[link](https://somedomain.org/the/path,"link, text")'

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "link",
                    "target": "https://somedomain.org/the/path",
                    "text": "link, text",
                }
            ],
        }
    ]

    _test(source, expected)


def test_macro_link_with_escaped_quotes():
    source = r'[link](https://somedomain.org/the/path,"link \"text\"")'

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "link",
                    "target": "https://somedomain.org/the/path",
                    "text": 'link "text"',
                }
            ],
        }
    ]

    _test(source, expected)


def test_automatic_link():
    source = "This is https://somedomain.org/the/path text"

    expected = [
        {
            "type": "sentence",
            "content": [
                {"type": "text", "value": "This is "},
                {
                    "type": "link",
                    "target": "https://somedomain.org/the/path",
                    "text": "https://somedomain.org/the/path",
                },
                {"type": "text", "value": " text"},
            ],
        }
    ]

    _test(source, expected)


def test_http_mentions():
    source = "Dangerous words: http and https"

    expected = [
        {
            "type": "sentence",
            "content": [
                {"type": "text", "value": "Dangerous words: http and https"},
            ],
        }
    ]

    _test(source, expected)


def test_macro_image():
    source = "[image](/the/path.jpg)"

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "image",
                    "uri": "/the/path.jpg",
                    "alt_text": None,
                    "width": None,
                    "height": None,
                }
            ],
        }
    ]

    _test(source, expected)


def test_macro_image_with_alt_text():
    source = '[image](/the/path.jpg,"alt name")'

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "image",
                    "uri": "/the/path.jpg",
                    "alt_text": "alt name",
                    "width": None,
                    "height": None,
                }
            ],
        }
    ]

    _test(source, expected)


def test_macro_image_with_width_and_height():
    source = "[image](/the/path.jpg,width=120,height=120)"

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "image",
                    "uri": "/the/path.jpg",
                    "alt_text": None,
                    "width": "120",
                    "height": "120",
                }
            ],
        }
    ]

    _test(source, expected)


@patch("mau.parsers.text_parser.footnote_anchors")
def test_macro_footnote(footnote_anchors_mock):
    footnote_anchors_mock.return_value = ("refXYZ", "defXYZ")

    source = "[footnote](Some text _and style_ and *more style* here)"

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "footnote_ref",
                    "number": 1,
                    "refanchor": "refXYZ",
                    "defanchor": "defXYZ",
                }
            ],
        }
    ]

    p = _test(source, expected)

    assert listasdict(p.footnotes) == [
        {
            "type": "footnote_def",
            "number": 1,
            "refanchor": "refXYZ",
            "defanchor": "defXYZ",
            "content": [
                {
                    "type": "sentence",
                    "content": [
                        {"type": "text", "value": "Some text "},
                        {
                            "type": "style",
                            "value": "underscore",
                            "content": {
                                "type": "sentence",
                                "content": [
                                    {"type": "text", "value": "and style"},
                                ],
                            },
                        },
                        {"type": "text", "value": " and "},
                        {
                            "type": "style",
                            "value": "star",
                            "content": {
                                "type": "sentence",
                                "content": [
                                    {"type": "text", "value": "more style"},
                                ],
                            },
                        },
                        {"type": "text", "value": " here"},
                    ],
                }
            ],
        }
    ]


@patch("mau.parsers.text_parser.footnote_anchors")
def test_parse_macro_footnote_can_start_with_different_number(footnote_anchors_mock):
    footnote_anchors_mock.return_value = ("refXYZ", "defXYZ")

    p = init_parser(
        "[footnote](Some text _and style_ and *more style* here)",
        footnotes_start_with=6,
    )
    result = p.parse()

    assert listasdict(p.nodes) == [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "footnote_ref",
                    "number": 6,
                    "refanchor": "refXYZ",
                    "defanchor": "defXYZ",
                }
            ],
        }
    ]


@patch("mau.parsers.text_parser.footnote_anchors")
def test_parse_macro_footnote_with_other_text_around(footnote_anchors_mock):
    footnote_anchors_mock.return_value = ("refXYZ", "defXYZ")

    source = "text[footnote](Some text) other text"

    expected = [
        {
            "type": "sentence",
            "content": [
                {"type": "text", "value": "text"},
                {
                    "type": "footnote_ref",
                    "number": 1,
                    "refanchor": "refXYZ",
                    "defanchor": "defXYZ",
                },
                {"type": "text", "value": " other text"},
            ],
        }
    ]

    p = _test(source, expected)

    assert listasdict(p.footnotes) == [
        {
            "type": "footnote_def",
            "number": 1,
            "refanchor": "refXYZ",
            "defanchor": "defXYZ",
            "content": [
                {
                    "type": "sentence",
                    "content": [
                        {"type": "text", "value": "Some text"},
                    ],
                }
            ],
        }
    ]


@patch("mau.parsers.text_parser.footnote_anchors")
def test_parse_macro_footnote_includes_quotes(footnote_anchors_mock):
    footnote_anchors_mock.return_value = ("refXYZ", "defXYZ")

    source = 'text[footnote]("Some text") other text'

    expected = [
        {
            "type": "sentence",
            "content": [
                {"type": "text", "value": "text"},
                {
                    "type": "footnote_ref",
                    "number": 1,
                    "refanchor": "refXYZ",
                    "defanchor": "defXYZ",
                },
                {"type": "text", "value": " other text"},
            ],
        }
    ]

    p = _test(source, expected)

    assert listasdict(p.footnotes) == [
        {
            "type": "footnote_def",
            "number": 1,
            "refanchor": "refXYZ",
            "defanchor": "defXYZ",
            "content": [
                {
                    "type": "sentence",
                    "content": [
                        {"type": "text", "value": '"Some text"'},
                    ],
                }
            ],
        }
    ]


@patch("mau.parsers.text_parser.footnote_anchors")
def test_parse_macro_footnote_includes_round_brakets(footnote_anchors_mock):
    footnote_anchors_mock.return_value = ("refXYZ", "defXYZ")

    source = r'text[footnote]("Some code(\)") other text'

    expected = [
        {
            "type": "sentence",
            "content": [
                {"type": "text", "value": "text"},
                {
                    "type": "footnote_ref",
                    "number": 1,
                    "refanchor": "refXYZ",
                    "defanchor": "defXYZ",
                },
                {"type": "text", "value": " other text"},
            ],
        }
    ]

    p = _test(source, expected)

    assert listasdict(p.footnotes) == [
        {
            "type": "footnote_def",
            "number": 1,
            "refanchor": "refXYZ",
            "defanchor": "defXYZ",
            "content": [
                {
                    "type": "sentence",
                    "content": [
                        {"type": "text", "value": '"Some code()"'},
                    ],
                }
            ],
        }
    ]


def test_parse_macro_mailto():
    source = "[mailto](info@projectmau.org)"

    expected = [
        {
            "type": "sentence",
            "content": [
                {
                    "type": "link",
                    "target": "mailto:info@projectmau.org",
                    "text": "info@projectmau.org",
                }
            ],
        }
    ]

    _test(source, expected)



from mau.lexers.base_lexer import Token, TokenTypes
from mau.lexers.text_lexer import TextLexer
from mau.parsers.base_parser import BaseParser, analyse, Literal
from mau.parsers.arguments_parser import ArgumentsParser, merge_args
from mau.parsers.nodes import (
    WordNode,
    SentenceNode,
    TextNode,
    StyleNode,
    VerbatimNode,
    ClassNode,
    MacroNode,
    LinkNode,
    ImageNode,
    FootnoteRefNode,
    FootnoteDefNode,
)

MAP_STYLES = {"_": "underscore", "*": "star"}


# Returns the id of the reference
# and the id of the definition
def footnote_anchors(content):
    h = str(hash(content))[:8]
    return f"fr-{h}", f"fd-{h}"  # pragma: no cover


class TextParser(BaseParser):
    def __init__(self, footnotes_start_with=1):
        super().__init__()

        self.lexer = TextLexer()

        self._styles = set()
        self._classes = set()
        self._verbatim = False

        self.footnotes_start_with = footnotes_start_with
        self.footnotes = []
        self._nodes = []
        self.nodes = []

    def parse_word(self):
        return WordNode(self.get_token_value())

    def parse_style(self):
        style = self.get_token_value(TokenTypes.LITERAL, check=lambda x: x in "*_")
        content = self.parse_sentence(stop_tokens={Token(TokenTypes.LITERAL, style)})
        self.get_token(TokenTypes.LITERAL, style)

        return StyleNode(MAP_STYLES[style], content)

    def parse_escape(self):
        self.get_token(TokenTypes.LITERAL, "\\")

        # if self.peek_token() == Literal('"'):
        #     return WordNode('\\"')

        return WordNode(self.get_token_value())

    def parse_styled_text(self, stop_tokens=None):
        if self.peek_token() in stop_tokens:
            return None

        with self:
            return self.parse_escape()

        with self:
            return self.parse_class()

        with self:
            return self.parse_macro()

        with self:
            return self.parse_verbatim()

        with self:
            return self.parse_style()

        with self:
            return self.parse_link()

        return self.parse_word()

    def parse_sentence(self, stop_tokens=None):
        content = []
        stop_tokens = stop_tokens or set()
        stop_tokens = stop_tokens.union({Token(TokenTypes.EOF), Token(TokenTypes.EOL)})

        result = self.parse_styled_text(stop_tokens)
        while result is not None:
            content.append(result)
            result = self.parse_styled_text(stop_tokens)

        import itertools

        # Group consecutive WordNode nodes into a single TextNode
        grouped_nodes = []
        for key, group in itertools.groupby(content, lambda x: x.__class__ == WordNode):
            if key:
                text = "".join([n.value for n in group])
                grouped_nodes.append(TextNode(text))
            else:
                grouped_nodes.extend(list(group))

        return SentenceNode(content=grouped_nodes)

    def parse_verbatim(self):
        self.get_token(TokenTypes.LITERAL, "`")
        content = self.collect_join(
            [Token(TokenTypes.LITERAL, "`"), Token(TokenTypes.EOL)],
        )
        self.get_token(TokenTypes.LITERAL, "`")

        return VerbatimNode(content)

    def parse_class(self):
        self.get_token(TokenTypes.LITERAL, "[")
        classes = self.collect_join(
            [Token(TokenTypes.LITERAL, "]"), Token(TokenTypes.EOL)]
        )
        self.get_token(TokenTypes.LITERAL, "]")
        self.get_token(TokenTypes.LITERAL, "#")
        content = self.parse_sentence(stop_tokens={Token(TokenTypes.LITERAL, "#")})
        self.get_token(TokenTypes.LITERAL, "#")

        classes = classes.split(",")

        return ClassNode(classes, content)

    def parse_macro_link(self, args, kwargs):
        args, kwargs = merge_args(args, kwargs, ["target", "text"])
        return LinkNode(kwargs.get("target"), kwargs.get("text"))

    def parse_macro_mailto(self, args, kwargs):
        args, kwargs = merge_args(args, kwargs, ["email"])
        email = kwargs.get("email")
        target = f"mailto:{email}"
        return LinkNode(target, email)

    def parse_macro_image(self, args, kwargs):
        args, kwargs = merge_args(
            args,
            kwargs,
            ["uri", "alt_text", "width", "height"],
        )

        uri = kwargs.pop("uri", None)
        alt_text = kwargs.pop("alt_text", None)
        width = kwargs.pop("width", None)
        height = kwargs.pop("height", None)

        return ImageNode(uri=uri, alt_text=alt_text, width=width, height=height)

    def parse_macro_footnote(self, args, kwargs):
        args, kwargs = merge_args(args, kwargs, ["content"])
        content_text = kwargs.get("content")
        refanchor, defanchor = footnote_anchors(content_text)
        number = self.footnotes_start_with + len(self.footnotes)

        p = analyse(TextParser(), content_text)

        self.footnotes.append(
            FootnoteDefNode(
                refanchor=refanchor, defanchor=defanchor, number=number, content=p.nodes
            )
        )
        return FootnoteRefNode(refanchor=refanchor, defanchor=defanchor, number=number)

    def parse_macro(self):
        self.get_token(TokenTypes.LITERAL, "[")
        macro_name = self.get_token_value(TokenTypes.TEXT)
        self.get_token(TokenTypes.LITERAL, "]")
        self.get_token(TokenTypes.LITERAL, "(")

        raw = False
        if macro_name == "footnote":
            raw = True

        arguments = self.collect_join(
            stop_tokens=[Token(TokenTypes.LITERAL, ")"), Token(TokenTypes.EOL)],
        )
        p = analyse(ArgumentsParser(raw=raw), arguments)

        self.get_token(TokenTypes.LITERAL, ")")

        if macro_name == "link":
            return self.parse_macro_link(args=p.args, kwargs=p.kwargs)
        if macro_name == "mailto":
            return self.parse_macro_mailto(args=p.args, kwargs=p.kwargs)
        elif macro_name == "image":
            return self.parse_macro_image(args=p.args, kwargs=p.kwargs)
        elif macro_name == "footnote":
            return self.parse_macro_footnote(args=p.args, kwargs=p.kwargs)

        return MacroNode(macro_name, args=p.args, kwargs=p.kwargs)

    def parse_link(self):
        link = self.get_token_value(
            TokenTypes.TEXT,
            check=lambda x: x.startswith("http://") or x.startswith("https://"),
        )

        return LinkNode(link, link)

    def _parse(self):
        self._save(self.parse_sentence())



import pytest

from mau.parsers.main_parser import MainParser, ParseError

from tests.helpers import init_parser_factory, parser_test_factory, dedent

init_parser = init_parser_factory(MainParser)

_test = parser_test_factory(MainParser)


def test_source():
    source = """
    [source,somelang]
    ----
    import os

    print(os.environ["HOME"])
    ----
    """

    expected = [
        {
            "type": "source",
            "language": "somelang",
            "callouts": {},
            "delimiter": ":",
            "kwargs": {},
            "title": None,
            "code": [
                {"type": "text", "value": "import os"},
                {"type": "text", "value": ""},
                {"type": "text", "value": 'print(os.environ["HOME"])'},
            ],
        },
    ]

    _test(source, expected)


def test_source_with_title():
    source = """
    . Title
    [source,somelang]
    ----
    import os

    print(os.environ["HOME"])
    ----
    """

    expected = [
        {
            "type": "source",
            "language": "somelang",
            "callouts": {},
            "delimiter": ":",
            "kwargs": {},
            "title": {
                "content": [{"type": "text", "value": "Title"}],
                "type": "sentence",
            },
            "code": [
                {"type": "text", "value": "import os"},
                {"type": "text", "value": ""},
                {"type": "text", "value": 'print(os.environ["HOME"])'},
            ],
        },
    ]

    _test(source, expected)


def test_source_without_language():
    source = """
            [source]
            ----
            import os

            print(os.environ["HOME"])
            ----
            """

    expected = [
        {
            "type": "source",
            "language": "text",
            "callouts": {},
            "delimiter": ":",
            "kwargs": {},
            "title": None,
            "code": [
                {"type": "text", "value": "import os"},
                {"type": "text", "value": ""},
                {"type": "text", "value": 'print(os.environ["HOME"])'},
            ],
        },
    ]

    _test(source, expected)


def test_source_ignores_mau_syntax():
    source = """
            [source]
            ----
            :answer:42
            The answer is {answer}
            ----
            """

    expected = [
        {
            "type": "source",
            "language": "text",
            "callouts": {},
            "delimiter": ":",
            "kwargs": {},
            "title": None,
            "code": [
                {"type": "text", "value": ":answer:42"},
                {"type": "text", "value": "The answer is {answer}"},
            ],
        },
    ]

    _test(source, expected)


def test_source_respects_spaces():
    source = """
            [source]
            ----
            // This is a comment
            ----
            """

    expected = [
        {
            "type": "source",
            "language": "text",
            "callouts": {},
            "delimiter": ":",
            "kwargs": {},
            "title": None,
            "code": [
                {"type": "text", "value": "// This is a comment"},
            ],
        },
    ]

    _test(source, expected)


def test_source_respects_indentation():
    source = """
            [source]
            ----
            * This is not indented
               ** This is indented
            ----
            """

    expected = [
        {
            "type": "source",
            "language": "text",
            "callouts": {},
            "delimiter": ":",
            "kwargs": {},
            "title": None,
            "code": [
                {"type": "text", "value": "* This is not indented"},
                {"type": "text", "value": "   ** This is indented"},
            ],
        },
    ]

    _test(source, expected)


def test_source_named_language():
    source = """
            [source,language=somelang]
            ----
            import os

            print(os.environ["HOME"])
            ----
            """

    expected = [
        {
            "type": "source",
            "language": "somelang",
            "callouts": {},
            "delimiter": ":",
            "kwargs": {},
            "title": None,
            "code": [
                {"type": "text", "value": "import os"},
                {"type": "text", "value": ""},
                {"type": "text", "value": 'print(os.environ["HOME"])'},
            ],
        },
    ]

    _test(source, expected)


def test_source_named_language_wins():
    source = """
            [source,text,language=somelang]
            ----
            import os

            print(os.environ["HOME"])
            ----
            """

    expected = [
        {
            "type": "source",
            "language": "somelang",
            "callouts": {},
            "delimiter": ":",
            "kwargs": {},
            "title": None,
            "code": [
                {"type": "text", "value": "import os"},
                {"type": "text", "value": ""},
                {"type": "text", "value": 'print(os.environ["HOME"])'},
            ],
        },
    ]

    _test(source, expected)


def test_source_callouts():
    source = """
            [source,somelang,callouts=":"]
            ----
            import sys
            import os:3:

            print(os.environ["HOME"]):6:
            ----
            3: This is an import
            6: Environment variables are paramount
            """

    expected = [
        {
            "type": "source",
            "language": "somelang",
            "callouts": {
                1: ("3", "This is an import"),
                3: ("6", "Environment variables are paramount"),
            },
            "delimiter": ":",
            "kwargs": {},
            "title": None,
            "code": [
                {"type": "text", "value": "import sys"},
                {"type": "text", "value": "import os"},
                {"type": "text", "value": ""},
                {"type": "text", "value": 'print(os.environ["HOME"])'},
            ],
        },
    ]

    _test(source, expected)


def test_source_callouts_possible_clash():
    source = """
            [source,somelang,callouts=":"]
            ----
            import sys
            import: os:3:

            print(os.environ["HOME"]):6:
            ----
            3: This is an import
            6: Environment variables are paramount
            """

    expected = [
        {
            "type": "source",
            "language": "somelang",
            "callouts": {
                1: ("3", "This is an import"),
                3: ("6", "Environment variables are paramount"),
            },
            "delimiter": ":",
            "kwargs": {},
            "title": None,
            "code": [
                {"type": "text", "value": "import sys"},
                {"type": "text", "value": "import: os"},
                {"type": "text", "value": ""},
                {"type": "text", "value": 'print(os.environ["HOME"])'},
            ],
        },
    ]

    _test(source, expected)
    _test(source, expected)


def test_source_callouts_default_delimiter():
    source = """
            [source,somelang]
            ----
            import sys
            import os:3:

            print(os.environ["HOME"]):6:
            ----
            3: This is an import
            6: Environment variables are paramount
            """

    expected = [
        {
            "type": "source",
            "language": "somelang",
            "callouts": {
                1: ("3", "This is an import"),
                3: ("6", "Environment variables are paramount"),
            },
            "delimiter": ":",
            "kwargs": {},
            "title": None,
            "code": [
                {"type": "text", "value": "import sys"},
                {"type": "text", "value": "import os"},
                {"type": "text", "value": ""},
                {"type": "text", "value": 'print(os.environ["HOME"])'},
            ],
        },
    ]

    _test(source, expected)


def test_source_callouts_custom_delimiter():
    source = """
            [source,somelang,callouts="|"]
            ----
            import sys
            import os|3|

            print(os.environ["HOME"])|6|
            ----
            3: This is an import
            6: Environment variables are paramount
            """

    expected = [
        {
            "type": "source",
            "language": "somelang",
            "callouts": {
                1: ("3", "This is an import"),
                3: ("6", "Environment variables are paramount"),
            },
            "delimiter": "|",
            "kwargs": {},
            "title": None,
            "code": [
                {"type": "text", "value": "import sys"},
                {"type": "text", "value": "import os"},
                {"type": "text", "value": ""},
                {"type": "text", "value": 'print(os.environ["HOME"])'},
            ],
        },
    ]

    _test(source, expected)


def test_source_callout_cannot_be_found():
    p = init_parser(
        dedent(
            """
            [source,somelang,callouts="|"]
            ----
            import sys
            import os

            print(os.environ["HOME"])|6|
            ----
            3: This is an import
            6: Environment variables are paramount
            """
        )
    )

    with pytest.raises(ParseError):
        p.parse()


def test_source_callouts_non_numeric_labels():
    source = """
            [source,somelang,callouts="|"]
            ----
            import sys
            import os|step1|

            print(os.environ["HOME"])|step2|
            ----
            step1: This is an import
            step2: Environment variables are paramount
            """

    expected = [
        {
            "type": "source",
            "language": "somelang",
            "callouts": {
                1: ("step1", "This is an import"),
                3: ("step2", "Environment variables are paramount"),
            },
            "delimiter": "|",
            "kwargs": {},
            "title": None,
            "code": [
                {"type": "text", "value": "import sys"},
                {"type": "text", "value": "import os"},
                {"type": "text", "value": ""},
                {"type": "text", "value": 'print(os.environ["HOME"])'},
            ],
        },
    ]

    _test(source, expected)


def test_source_callout_wrong_format():
    p = init_parser(
        dedent(
            """
            [source,somelang,callouts=":"]
            ----
            import sys
            import os:3:

            print(os.environ["HOME"]):6:
            ----
            3 This is an import
            6: Environment variables are paramount
            """
        )
    )

    with pytest.raises(ParseError):
        p.parse()



from unittest.mock import patch

from mau.parsers.main_parser import MainParser

from tests.helpers import listasdict, init_parser_factory, parser_test_factory, dedent

init_parser = init_parser_factory(MainParser)

_test = parser_test_factory(MainParser)


@patch("mau.parsers.main_parser.header_anchor")
@patch("mau.parsers.text_parser.footnote_anchors")
def test_macro_footnote(footnote_anchors_mock, header_anchor_mock):
    footnote_anchors_mock.return_value = ("refXYZ", "defXYZ")
    header_anchor_mock.return_value = "HXXXXXX"

    source = """
    Some text[footnote](just a note)
    """

    expected = [
        {
            "type": "paragraph",
            "args": [],
            "kwargs": {},
            "content": {
                "type": "sentence",
                "content": [
                    {
                        "type": "text",
                        "value": "Some text",
                    },
                    {
                        "refanchor": "refXYZ",
                        "defanchor": "defXYZ",
                        "number": 1,
                        "type": "footnote_ref",
                    },
                ],
            },
        },
    ]

    p = _test(source, expected)

    assert listasdict(p.footnotes) == [
        {
            "type": "footnote_def",
            "refanchor": "refXYZ",
            "defanchor": "defXYZ",
            "number": 1,
            "content": [
                {
                    "type": "sentence",
                    "content": [
                        {"type": "text", "value": "just a note"},
                    ],
                }
            ],
        }
    ]


@patch("mau.parsers.text_parser.footnote_anchors")
def test_macro_footnote_inside_admonition(footnote_anchors_mock):
    footnote_anchors_mock.return_value = ("refXYZ", "defXYZ")

    p = init_parser(
        dedent(
            """
            Some text[footnote](note number 1)

            [admonition, someclass, someicon, somelabel]
            ----
            Some text[footnote](note number 2)
            ----
            """
        )
    )
    p.parse()

    assert listasdict(p.footnotes) == [
        {
            "type": "footnote_def",
            "refanchor": "refXYZ",
            "defanchor": "defXYZ",
            "number": 1,
            "content": [
                {
                    "type": "sentence",
                    "content": [
                        {"type": "text", "value": "note number 1"},
                    ],
                }
            ],
        },
        {
            "type": "footnote_def",
            "refanchor": "refXYZ",
            "defanchor": "defXYZ",
            "number": 1,
            "content": [
                {
                    "type": "sentence",
                    "content": [
                        {"type": "text", "value": "note number 2"},
                    ],
                }
            ],
        },
    ]



from unittest.mock import patch

from mau.parsers.main_parser import MainParser

from tests.helpers import listasdict, init_parser_factory, parser_test_factory, dedent

init_parser = init_parser_factory(MainParser)

_test = parser_test_factory(MainParser)


@patch("mau.parsers.main_parser.header_anchor")
def test_create_toc(header_anchor_mock):
    header_anchor_mock.side_effect = lambda text, level: f"{text}-XXXXXX"

    p = init_parser(
        dedent(
            """
            = Header 1
            == Header 1.1
            == Header 1.2
            = Header 2
            == Header 2.1
            === Header 2.1.1
            """
        )
    )
    p.parse()

    assert listasdict(p.toc) == [
        {
            "type": "toc_entry",
            "level": 1,
            "text": "Header 1",
            "anchor": "Header 1-XXXXXX",
            "children": [
                {
                    "type": "toc_entry",
                    "level": 2,
                    "text": "Header 1.1",
                    "anchor": "Header 1.1-XXXXXX",
                    "children": [],
                },
                {
                    "type": "toc_entry",
                    "level": 2,
                    "text": "Header 1.2",
                    "anchor": "Header 1.2-XXXXXX",
                    "children": [],
                },
            ],
        },
        {
            "type": "toc_entry",
            "level": 1,
            "text": "Header 2",
            "anchor": "Header 2-XXXXXX",
            "children": [
                {
                    "type": "toc_entry",
                    "level": 2,
                    "text": "Header 2.1",
                    "anchor": "Header 2.1-XXXXXX",
                    "children": [
                        {
                            "type": "toc_entry",
                            "level": 3,
                            "text": "Header 2.1.1",
                            "anchor": "Header 2.1.1-XXXXXX",
                            "children": [],
                        },
                    ],
                },
            ],
        },
    ]


@patch("mau.parsers.main_parser.header_anchor")
def test_create_toc_ignore_headers(header_anchor_mock):
    header_anchor_mock.side_effect = lambda text, level: f"{text}-XXXXXX"

    p = init_parser(
        dedent(
            """
            = Header 1
            ==! Header 1.1
            ==! Header 1.2
            = Header 2
            ==! Header 2.1
            ===! Header 2.1.1
            """
        )
    )
    p.parse()

    assert listasdict(p.toc) == [
        {
            "type": "toc_entry",
            "level": 1,
            "text": "Header 1",
            "anchor": "Header 1-XXXXXX",
            "children": [],
        },
        {
            "type": "toc_entry",
            "level": 1,
            "text": "Header 2",
            "anchor": "Header 2-XXXXXX",
            "children": [],
        },
    ]


@patch("mau.parsers.main_parser.header_anchor")
def test_create_toc_orphan_nodes(header_anchor_mock):
    header_anchor_mock.side_effect = lambda text, level: f"{text}-XXXXXX"

    p = init_parser(
        dedent(
            """
            = Header 1
            === Header 1.1.1
            == Header 1.2
            """
        )
    )
    p.parse()

    assert [i.asdict() for i in p.toc] == [
        {
            "type": "toc_entry",
            "level": 1,
            "text": "Header 1",
            "anchor": "Header 1-XXXXXX",
            "children": [
                {
                    "type": "toc_entry",
                    "level": 3,
                    "text": "Header 1.1.1",
                    "anchor": "Header 1.1.1-XXXXXX",
                    "children": [],
                },
                {
                    "type": "toc_entry",
                    "level": 2,
                    "text": "Header 1.2",
                    "anchor": "Header 1.2-XXXXXX",
                    "children": [],
                },
            ],
        },
    ]


@patch("mau.parsers.main_parser.header_anchor")
def test_parse_header_not_in_toc(header_anchor_mock):
    header_anchor_mock.return_value = "XXXXXX"

    source = """
    =! Title of the section
    """

    expected = [
        {
            "type": "header",
            "kwargs": {},
            "value": "Title of the section",
            "level": 1,
            "anchor": "XXXXXX",
        }
    ]

    p = _test(source, expected)

    assert listasdict(p.toc) == []



import functools

from mau.lexers.base_lexer import BaseLexer, Token, TokenTypes, TokenError, Literal


class ParseError(ValueError):
    pass


class ExpectedError(ParseError):
    pass


# This makes a function return True instead of
# None when successfully executed, which
# makes the parser loop much easier to
# understand
def parser(func):
    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs):
        value = func(*args, **kwargs)
        if value is None:
            return True

    return wrapper_decorator


def analyse(parser, text):
    parser.load(text)
    parser.parse()

    return parser


class BaseParser:
    def __init__(self):
        self.lexer = BaseLexer()
        self.index = -1
        self.tokens = []
        self._stack = []
        self.nodes = []

    def _stash(self):
        self._stack.append(self.index)

    def __enter__(self):
        self._stash()

    def __exit__(self, etype, evalue, etrace):
        # Leave the stack as it was before
        # we entered the context
        stacked_index = self._pop()

        if etype:
            # Restore the original position
            self.index = stacked_index

        if etype in [TokenError]:
            return True

    def _pop(self):
        return self._stack.pop()

    def fail(self):
        raise TokenError

    def _save(self, node):
        self.nodes.append(node)

    @property
    def current_token(self):
        try:
            return self.tokens[self.index]
        except IndexError:
            return Token(TokenTypes.EOF)

    def load(self, text):
        self.lexer.process(text)
        self.tokens = self.lexer.tokens
        self.index = -1

    def _check_token(self, token, ttype=None, tvalue=None, check=None):
        check_type = ttype if ttype is not None else token.type

        if token.type != check_type:
            raise TokenError(f"Type of token {token} is not {check_type}")

        if tvalue is not None and token.value != tvalue:
            raise TokenError(f"Value of token {token} is not {tvalue}")
        elif check is not None and check(token.value) is False:
            raise TokenError(f"Value of token {token} didn't pass check")

        return token

    def get_token(self, ttype=None, tvalue=None, check=None):
        """
        Return the next token and advances the index.
        The token is stored it in current_token.
        """

        if self.index == len(self.tokens):
            return Token(TokenTypes.EOF)

        self.index += 1
        return self._check_token(self.current_token, ttype, tvalue, check)

    def get_token_value(self, ttype=None, tvalue=None, check=None):
        """
        A shortcut to get the next token and return the value
        """

        return self.get_token(ttype, tvalue, check).value

    def check_current_token(self, ttype, tvalue=None, check=None):
        """
        Just check the type and value of the current token without
        advancing the index.
        """

        return self._check_token(self.current_token, ttype, tvalue, check)

    def peek_token(self, ttype=None, tvalue=None, check=None):
        """
        Return the next token without advancing the index.
        """

        try:
            token = self.tokens[self.index + 1]
            return self._check_token(token, ttype, tvalue, check)
        except IndexError:
            return Token(TokenTypes.EOF)

    def peek_token_is(self, *args, **kwargs):
        try:
            self.peek_token(*args, **kwargs)
            return True
        except TokenError:
            return False

    def force_token(self, ttype, tvalue=None):
        try:
            return self.get_token(ttype, tvalue)
        except TokenError:
            raise ExpectedError(
                {"expected": Token(ttype, tvalue), "found": self.current_token}
            )

    def force_token_in(self, tokens):
        if self.peek_token() not in tokens:
            raise ExpectedError({"expected": tokens, "found": self.current_token})

    def collect(self, stop_tokens, preserve_escaped_stop_tokens=False):
        tokens = []

        while self.peek_token() not in stop_tokens:
            if self.peek_token() == Literal("\\"):
                escape = self.get_token()

                if self.peek_token() not in stop_tokens or preserve_escaped_stop_tokens:
                    tokens.append(escape)

            tokens.append(self.get_token())

        return tokens

    def collect_join(
        self, stop_tokens, join_with="", preserve_escaped_stop_tokens=False
    ):
        token_values = [
            t.value for t in self.collect(stop_tokens, preserve_escaped_stop_tokens)
        ]
        token_values = [t for t in token_values if t is not None]

        return join_with.join(token_values)

    def parse(self):
        self._parse()

    def _parse(self):
        while not self.peek_token_is(TokenTypes.EOF):
            result = False

            for parse_function in self._parse_functions():
                with self:
                    result = parse_function()

                # We get here both if the context manager
                # was interrupted by the exception or if
                # it finished properly. In this last case,
                # however, the function reached the end
                # and returned True
                if result is True:
                    break

            # If we get here and result is still false
            # we didn't find any function to parse the
            # current token.
            if result is False:
                raise ParseError(f"Cannot parse token {self.peek_token()}")



# There are two types of elements in Mau, inline and block ones
# (the nomenclature is borrowed from HTML).
# Block elements are always on a new line, so they correspond to
# main items in the document, while inline elements can only appear
# inside a paragraph.


class Node:
    node_type = "node"

    def asdict(self):
        return {"type": self.node_type}  # pragma: no cover


#############################################################
# Inline elements


class ValueNode(Node):
    node_type = "value_node"

    def __init__(self, value):
        self.value = value

    def asdict(self):
        return {"type": self.node_type, "value": self.value}


class ContainerNode(Node):
    node_type = "container_node"

    def __init__(self, content):
        self.content = content

    def asdict(self):
        return {"type": self.node_type, "content": [i.asdict() for i in self.content]}


class WrapperNode(ValueNode):
    node_type = "wrapper_node"

    def __init__(self, value, content):
        super().__init__(value)
        self.content = content

    def asdict(self):
        return {
            "type": self.node_type,
            "value": self.value,
            "content": self.content.asdict(),
        }


class WordNode(ValueNode):
    """This is a single word, it's used internally
    and eventually packed together with others into
    a TextNode
    """

    node_type = "word"


class TextNode(ValueNode):
    """This contains plain text and is created
    as a collation of multiple WordNode objects

    Attributes:
        value: the text contained in the node
    """

    node_type = "text"


class SentenceNode(ContainerNode):
    """A recursive container node.

    This node represents the content of a paragraph, but it is recursive,
    while :obj:`ParagraphNode` is not. It can contain other :obj:`SentenceNode`
    objects, or other types of nodes like :obj:`TextNode`,
    :obj:`StyleNode` or :obj:`MacroNode`.

    Attributes:
        content (:obj:`list`): a list of nodes contained in this node
    """

    node_type = "sentence"


class StyleNode(WrapperNode):
    """Describes the style applied to a node.

    It contains a single node that is styled with the given value.
    Currently Mau supports 2 styles represented by `*` and `_`
    but this node is agnostic.

    Attributes:
        value (:obj:`str`): the name of the style
        content (:obj:`SentenceNode`): a single node
    """

    node_type = "style"


class VerbatimNode(ValueNode):
    """Verbatim text.

    This node contains verbatim text.

    Attributes:
        value: the text contained in the node
    """

    node_type = "verbatim"


class ClassNode(Node):
    """Text with one or more class.

    This node contains text that has been assigned
    one or more classes.

    Attributes:
        value: the text contained in the node
        classes (:obj:`list`): the classes assigned to the text
    """

    node_type = "class"

    def __init__(self, classes, content):
        self.classes = classes
        self.content = content

    def asdict(self):
        return {
            "type": self.node_type,
            "classes": self.classes,
            "content": self.content.asdict(),
        }


class MacroNode(Node):
    """Base node for macros.

    This node contains a macro, with a name and arguments.

    Attributes:
        value: the name of the macro
        args (:obj:`tuple`): the unnamed arguments
        kwargs (:obj:`dict`): the named arguments
    """

    node_type = "macro"

    def __init__(self, value, args=None, kwargs=None):
        self.value = value
        self.args = args or []
        self.kwargs = kwargs or {}

    def asdict(self):
        return {
            "type": self.node_type,
            "value": self.value,
            "args": self.args,
            "kwargs": self.kwargs,
        }


class LinkNode(Node):
    """Link macro.

    This node contains a link.

    Attributes:
        target: the target of the link
        text: the text of the link
    """

    node_type = "link"

    def __init__(self, target, text):
        self.target = target
        self.text = text

    def asdict(self):
        return {
            "type": self.node_type,
            "target": self.target,
            "text": self.text,
        }


class ImageNode(Node):
    """Inline image.

    This node contains an inline image.

    Attributes:
        uri: the URI of the image
        alt_text: alternative text if the image is not available
        width: width of the image
        height: height fo the image
    """

    node_type = "image"

    def __init__(self, uri, alt_text=None, width=None, height=None):
        self.uri = uri
        self.alt_text = alt_text
        self.width = width
        self.height = height

    def asdict(self):
        return {
            "type": self.node_type,
            "uri": self.uri,
            "alt_text": self.alt_text,
            "width": self.width,
            "height": self.height,
        }


class FootnoteRefNode(Node):
    """Footnote reference.

    This node contains a footnote reference.

    Attributes:
       number: number of this footnote
       refanchor: anchor ID of the reference
       defanchor: anchor ID of the definition
    """

    node_type = "footnote_ref"

    def __init__(self, number, refanchor, defanchor):
        self.number = number
        self.refanchor = refanchor
        self.defanchor = defanchor

    def asdict(self):
        return {
            "type": self.node_type,
            "refanchor": self.refanchor,
            "defanchor": self.defanchor,
            "number": self.number,
        }


#############################################################
# Block elements


class HorizontalRuleNode(Node):
    """A horizontal rule."""

    def asdict(self):
        return {"type": "horizontal_rule"}


class ContentImageNode(Node):
    """An image included in the page.

    This represents an image included in the page.

    Arguments:
        uri: the URI of the image
        alt_text: alternative text if the image is not available
        classes (:obj:`list`): list of classes added to the image
        title: caption of the image
        kwargs: named arguments
    """

    node_type = "content_image"

    def __init__(self, uri, alt_text, classes=None, title=None, kwargs=None):
        self.uri = uri
        self.alt_text = alt_text
        self.classes = classes
        self.title = title
        self.kwargs = kwargs or {}

    def asdict(self):
        return {
            "type": self.node_type,
            "alt_text": self.alt_text,
            "uri": self.uri,
            "title": self.title.asdict() if self.title else None,
            "classes": self.classes,
            "kwargs": self.kwargs,
        }


class ParagraphNode(Node):
    """A paragraph.

    This is a wrapper around a single :obj:`SentenceNode`. This node
    is not recursive.

    Arguments:
        content (:obj:`SentenceNode`): the text of the paragraph
        args: unnamed arguments
        kwargs: named arguments
    """

    node_type = "paragraph"

    def __init__(self, content, args=None, kwargs=None):
        self.args = args or []
        self.kwargs = kwargs or {}
        self.content = content

    def asdict(self):
        return {
            "type": self.node_type,
            "content": self.content.asdict(),
            "args": self.args,
            "kwargs": self.kwargs,
        }


class HeaderNode(Node):
    """A header.

    This node represents a header.

    Arguments:
        value: the text in the header
        level: the level of the header (1 is most important)
        anchor: the ID anchor of this node
        kwargs: named arguments
    """

    node_type = "header"

    def __init__(self, value, level, anchor, kwargs=None):
        self.value = value
        self.level = level
        self.anchor = anchor
        self.kwargs = kwargs or {}

    def asdict(self):
        return {
            "type": self.node_type,
            "value": self.value,
            "level": self.level,
            "anchor": self.anchor,
            "kwargs": self.kwargs,
        }


class BlockNode(Node):
    """A block.

    This node contains a generic block.

    Arguments:
        blocktype: the type of this block
        content: content of the block
        secondary_content: secondary content of this block
        title: title of this block
        args: unnamed arguments
        kwargs: named arguments
    """

    node_type = "block"

    def __init__(
        self, blocktype, content, secondary_content, title=None, args=None, kwargs=None
    ):
        self.blocktype = blocktype
        self.content = content
        self.secondary_content = secondary_content
        self.title = title
        self.args = args or []
        self.kwargs = kwargs or {}

    def asdict(self):
        return {
            "type": self.node_type,
            "blocktype": self.blocktype,
            "content": [i.asdict() for i in self.content],
            "secondary_content": [i.asdict() for i in self.secondary_content],
            "args": self.args,
            "kwargs": self.kwargs,
            "title": self.title.asdict() if self.title else None,
        }


class SourceNode(Node):
    """A block of verbatim text or source code.

    This node contains verbatim text or source code.

    Arguments:
        language: the language of the code contained in this block
        callouts: callouts for this source code
        delimiter: callouts delimiter
        code: content of the block
        title: title of this block
        kwargs: named arguments
    """

    node_type = "source"

    def __init__(self, language, callouts, delimiter, code, title=None, kwargs=None):
        self.language = language
        self.callouts = callouts
        self.delimiter = delimiter
        self.code = code
        self.title = title
        self.kwargs = kwargs or {}

    def asdict(self):
        return {
            "type": self.node_type,
            "language": self.language,
            "callouts": self.callouts,
            "delimiter": self.delimiter,
            "code": [i.asdict() for i in self.code],
            "title": self.title.asdict() if self.title else None,
            "kwargs": self.kwargs,
        }


class RawNode(Node):
    """A block of raw content.

    The content of this node will be inserted exactly as it is.

    Arguments:
        content: the content of the block
        args: unnamed arguments
        kwargs: named arguments
    """

    node_type = "raw"

    def __init__(self, content, args=None, kwargs=None):
        self.content = content
        self.args = args or []
        self.kwargs = kwargs or {}

    def asdict(self):
        return {
            "type": self.node_type,
            "content": [i.asdict() for i in self.content],
            "args": self.args,
            "kwargs": self.kwargs,
        }


class AdmonitionNode(Node):
    """An admonition.

    An admonition node that has a class, an icon, a label, and content.

    Arguments:
        admclass: the class of the admonition
        icon: the icon assigned to this admonition
        label: a label that acts as a title for the admonition
        content: the content of the admonition
        kwargs: named arguments
    """

    node_type = "admonition"

    def __init__(self, admclass, icon, label, content, kwargs=None):
        self.admclass = admclass
        self.icon = icon
        self.label = label
        self.content = content
        self.kwargs = kwargs or {}

    def asdict(self):
        return {
            "type": self.node_type,
            "class": self.admclass,
            "icon": self.icon,
            "label": self.label,
            "content": [i.asdict() for i in self.content],
            "kwargs": self.kwargs,
        }


class CommandNode(Node):
    """A Mau command.

    This represents a command issued in a Mau document.

    Arguments:
        name: the name of the command
        args: unnamed arguments
        kwargs: named arguments
    """

    node_type = "command"

    def __init__(self, name, args=None, kwargs=None):
        self.kwargs = kwargs or {}
        self.name = name

    def asdict(self):
        return {
            "type": self.node_type,
            "name": self.name,
            "kwargs": self.kwargs,
        }


class QuoteNode(Node):
    """A quote.

    This node contains a quote with text and an attribution.

    Arguments:
        attribution: the attribution for this quote
        content: the quote itself
        kwargs: named arguments
    """

    node_type = "quote"

    def __init__(self, attribution, content, kwargs=None):
        self.attribution = attribution
        self.content = content
        self.kwargs = kwargs or {}

    def asdict(self):
        return {
            "type": self.node_type,
            "attribution": self.attribution,
            "content": [i.asdict() for i in self.content],
            "kwargs": self.kwargs,
        }


class FootnoteDefNode(Node):
    """A footnote definition.

    This node contains the definition of a footnote(i.e. the full text)

    Arguments:
        number: number of this footnote
        refanchor: the ID of the reference of this footnote
        defanchor: the ID of this definition
        content: the full text of the footnote
    """

    node_type = "footnote_def"

    def __init__(self, number, refanchor, defanchor, content):
        self.number = number
        self.refanchor = refanchor
        self.defanchor = defanchor
        self.content = content

    def asdict(self):
        return {
            "type": self.node_type,
            "refanchor": self.refanchor,
            "defanchor": self.defanchor,
            "number": self.number,
            "content": [i.asdict() for i in self.content],
        }


class TocNode(Node):
    """An entry of the Table of Contents.

    This node contains an entry of the Table of Contents.

    Arguments:
        level: level of this entry (level of the header)
        text: the text of the entry
        anchor: the ID of the header
        children: child entries
    """

    node_type = "toc_entry"

    def __init__(self, level, text, anchor):
        self.level = level
        self.text = text
        self.anchor = anchor
        self.children = []

    def asdict(self):
        return {
            "type": self.node_type,
            "level": self.level,
            "text": self.text,
            "anchor": self.anchor,
            "children": [i.asdict() for i in self.children],
        }


class ListNode(Node):
    """A list.

    This node contains an ordered or unordered list.

    Arguments:
        ordered (:obj:`bool`): whether the list is ordered or not
        items (:obj:`list`): entries contained in this entry
        main_node (:obj:`bool`): whether this is the main node of the list
    """

    node_type = "list"

    def __init__(self, ordered, items, main_node=False):
        self.ordered = ordered
        self.items = items
        self.main_node = main_node

    def asdict(self):
        return {
            "type": self.node_type,
            "ordered": self.ordered,
            "items": [i.asdict() for i in self.items],
            "main_node": self.main_node,
        }


class ListItemNode(Node):
    """An entry in a list.

    This node contains an entry of an ordered or unordered list.

    Arguments:
        level: the level of depth of this entry in the list
        content: the content of the entry
    """

    node_type = "list_item"

    def __init__(self, level, content):
        self.level = level
        self.content = content

    def asdict(self):
        return {
            "type": self.node_type,
            "level": self.level,
            "content": self.content.asdict(),
        }


class DocumentNode(Node):
    """A document.

    This node represents the full document.

    Arguments:
        content: the content of the document
        no_document (:obj:`bool`): if True the content won't be wrapped by a template
    """

    node_type = "document"

    def __init__(self, content, no_document=False):
        self.content = content
        self.no_document = no_document

    def asdict(self):
        return {
            "type": self.node_type,
            "no_document": self.no_document,
            "content": [i.asdict() for i in self.content],
        }
